from __future__ import annotations

"""
Stoplight OpenAPI Docs Params
"""

from lazyops.types import BaseModel, lazyproperty
from fastapi.responses import HTMLResponse
from typing import Optional, Dict, Any


class OpenAPIStoplight(BaseModel):
    app_title: Optional[str] = ''
    version: Optional[str] = 'latest'
    style: Optional[str] = ''
    favicon_url: Optional[str] = ''
    logo_url: Optional[str] = ''
    openapi_url: Optional[str] = '/openapi.json'
    google_analytics_tag_id: Optional[str] = None


    @lazyproperty
    def css_url(self):
        """
        Return the CSS URL for the Stoplight Elements.
        """
        return f"https://unpkg.com/@stoplight/elements@{self.version}/styles.min.css" \
            if (self.version and self.version != "latest") else "https://unpkg.com/@stoplight/elements/styles.min.css"
    
    @lazyproperty
    def js_url(self):
        """
        Return the JS URL for the Stoplight Elements.
        """
        return f"https://unpkg.com/@stoplight/elements@{self.version}/web-components.min.js" \
            if (self.version and self.version != "latest") else "https://unpkg.com/@stoplight/elements/web-components.min.js"
    

    @property
    def favicon(self) -> str:
        """Return favicon `<link>` tag, if applicable.
        Returns:
            A `<link>` tag if self.favicon_url is not empty, otherwise returns a placeholder meta tag.
        """
        return f"<link rel='icon' type='image/x-icon' href='{self.favicon_url}'>" if self.favicon_url else "<meta/>"


    @property
    def logo_data(self) -> str:
        """
        Return logo data
        """
        return f'logo="{self.logo_url}"' if self.logo_url else ''
    
    @lazyproperty
    def html_body(self):
        """
        Return the HTML body
        """
        head = f"""
            <head>
            <title>{self.app_title} - Documentation</title>
            {self.favicon}
            <meta charset="utf-8"/>
            <meta name="viewport" content="width=device-width, initial-scale=1, shrink-to-fit=no">
            <link rel="stylesheet" href="{self.css_url}">
            <script src="{self.js_url}" crossorigin></script>
            <style>{self.style}</style>
            </head>
        """
        if self.google_analytics_tag_id:
            head += """
        <!-- Google tag (gtag.js) -->
        <script async src="https://www.googletagmanager.com/gtag/js?id=<GOOGLE_TAG_ID>"></script>
        <script>
            window.dataLayer = window.dataLayer || [];
            function gtag(){dataLayer.push(arguments);}
            gtag('js', new Date());

            gtag('config', '<GOOGLE_TAG_ID>');
        </script>    
        """.replace('<GOOGLE_TAG_ID>', self.google_analytics_tag_id)

        _set_header_script = """
            <script>
            // Check whether `x-api-key` is present in the headers
            // and set `TryIt_securitySchemeValues` value in localStorage accordingly
            // const apiKey = localStorage.getItem('x-api-key');
            function parseHttpHeaders(httpHeaders) {
                return httpHeaders.split("\\n")
                .map(x=>x.split(/: */,2))
                .filter(x=>x[0])
                .reduce((ac, x)=>{ac[x[0]] = x[1];return ac;}, {});
            }

            var req = new XMLHttpRequest();
            req.open('GET', document.location, false);
            req.send(null);
            var headers = parseHttpHeaders(req.getAllResponseHeaders());

            // const apiKey = Headers().get('x-api-key');
            const apiKey = headers['x-api-key'];
            const authHeader = headers['authorization'];
            if (apiKey) {
                localStorage.setItem('TryIt_securitySchemeValues', JSON.stringify({
                    'API Key': apiKey.trim(),
                    'Auth0 Authorization': authHeader ? authHeader.trim() : ''
                }));
            }
            // else if (authHeader) {
            //     localStorage.setItem('TryIt_securitySchemeValues', JSON.stringify({       
            // }
            // if (authHeader) {
            
            </script>
        """

        body = f"""
            <body>
            {_set_header_script}
            <elements-api
                apiDescriptionUrl="{self.openapi_url}"
                router="hash"
                layout="sidebar"
                tryItCredentialsPolicy="include"
                {self.logo_data}
            />
            </body>
        """

        return f"""
        <!DOCTYPE html>
            <html>
                {head}
                {body}
            </html>
        """.strip()

    def get_response(self, headers: Optional[Dict[str, Any]] = None, **kwargs) -> HTMLResponse:
        """
        Return the HTML response
        """
        return HTMLResponse(content=self.html_body, headers=headers, **kwargs)


from __future__ import annotations

import os
import abc
import atexit

import pathlib
import filelock
import contextlib
from lazyops.types import BaseModel, Field
from lazyops.utils.logs import logger
from lazyops.utils.serialization import Json
from typing import Optional, Dict, Any, Set, List, Union, TYPE_CHECKING

with contextlib.suppress(ImportError):
    from fastapi import FastAPI

if TYPE_CHECKING:
    from lazyops.types.models import BaseSettings


class StateData(abc.ABC):

    def __init__(self, filepath: pathlib.Path):
        self.filepath = filepath
        self.filelock_path = filepath.with_suffix('.lock')
        self.filelock = filelock.FileLock(lock_file = self.filelock_path.as_posix(), thread_local = False)
        with self.filelock:
            if not self.filepath.exists():
                self.filepath.write_text('{}')

    @property
    def data(self) -> Dict[str, Any]:
        """
        Returns the data
        """
        with self.filelock:
            return Json.loads(self.filepath.read_text())
        
    @contextlib.contextmanager
    def ctx(self) -> Dict[str, Union[List[Any], Dict[str, Any], Any]]:
        """
        Returns the context
        """
        with self.filelock:
            try:
                data = self.data
                yield data
            finally:
                self.filepath.write_text(Json.dumps(data))
    
    def get(self, key: str, default: Optional[Any] = None) -> Any:
        """
        Returns the value for the given key
        """
        return self.data.get(key, default)
    
    def __contains__(self, key: str) -> bool:
        """
        Returns whether the key is in the data
        """
        return key in self.data
    
    def __getitem__(self, key: str) -> Any:
        """
        Returns the value for the given key
        """
        return self.data.get(key)
    
    def __setitem__(self, key: str, value: Any):
        """
        Sets the value for the given key
        """
        with self.ctx() as data:
            data[key] = value

    def __delitem__(self, key: str):
        """
        Deletes the value for the given key
        """
        with self.ctx() as data:
            del data[key]


    def __iter__(self):
        """
        Returns the iterator
        """
        return iter(self.data)
    
    def __len__(self) -> int:
        """
        Returns the length
        """
        return len(self.data)
    
    def __repr__(self) -> str:
        """
        Returns the representation
        """
        return repr(self.data)
    
    def __str__(self) -> str:
        """
        Returns the string representation
        """
        return str(self.data)
    
    def __bool__(self) -> bool:
        """
        Returns whether the data is empty
        """
        return bool(self.data)
    
    def __eq__(self, other: Any) -> bool:
        """
        Returns whether the data is equal to the other
        """
        return self.data == other
    
    def keys(self) -> Set[str]:
        """
        Returns the keys
        """
        return self.data.keys()
    
    def setdefault(self, key: str, default: Any) -> Any:
        """
        Sets the default value for the given key
        """
        with self.ctx() as data:
            value = data.setdefault(key, default)
        return value
        
    def close(self):
        """
        Closes the filelock
        """
        self.filelock.release()

    def append(self, key: str, value: Any) -> bool:
        """
        Appends the value to the list
        """
        with self.ctx() as data:
            if key not in data:
                data[key] = []
            if value not in data[key]:
                data[key].append(value)
                return False
        return True


StateData.register(dict)


class AppState(BaseModel):
    """
    Holds the state of the current settings
    """
    configured: Set[str] = Field(default_factory=set)
    initialized: Set[str] = Field(default_factory=set)
    completed: bool = False

    api_app: Optional["FastAPI"] = None

    ctx: Optional[Dict[str, Any]] = {}

    app_entrypoint: Optional[str] = None
    worker_entrypoint: Optional[str] = None

    k8s_kubeconfigs: Optional[Dict[str, str]] = Field(default_factory=dict)
    k8s_active_ctx: Optional[str] = None

    app_module_name: Optional[str] = None
    data_path: Optional[pathlib.Path] = None

    class Config:
        extra = 'allow'
        arbitrary_types_allowed = True

    @property
    def is_silent(self) -> bool:
        """
        Returns whether the current state is silent
        """
        return self.ctx.get('silent', False)
    
    @property
    def settings(self) -> 'BaseSettings':
        """
        Returns the settings
        """
        return self.ctx.get('settings', None)

    @property
    def stx(self) -> Optional['StateData']:
        """
        Returns the StateData
        """
        return self.ctx.get('stx', None)

    @property
    def process_id(self) -> Optional[int]:
        """
        Returns the process id
        """
        return self.ctx.get('process_id', None)
    
    @property
    def is_primary_process(self) -> bool:
        """
        Returns whether this is the primary process
        """
        return self.process_id == self.stx.get('primary_process_id', 0)
    
    @property
    def is_primary_server_process(self) -> bool:
        """
        Returns whether this is the primary server process
        """
        return self.process_id == self.stx.get('primary_server_process_id', 0)

    @property
    def is_leader_process(self) -> bool:
        """
        Returns whether this is the leader process
        """
        return self.process_id in self.stx.get('leader_process_ids', []) or self.is_primary_process

    
    @property
    def server_process_id_path(self) -> pathlib.Path:
        """
        Returns the server process id path
        """
        return self.data_path.joinpath(f'{self.app_module_name}.pid')

    def configure_silent(self, _silent: Optional[bool] = None):
        """
        Configures the silent mode
        """
        if _silent is not None:
            self.ctx['silent'] = _silent

    def get_kubeconfig(
        self, 
        name: Optional[str] = None, 
        set_as_envval: Optional[bool] = True,
        set_active: Optional[bool] = False,
    ) -> str:
        """
        Returns the kubeconfig for the given context
        """
        name = name or self.k8s_active_ctx
        if name is not None and name in self.k8s_kubeconfigs:
            kconfig = self.k8s_kubeconfigs[name]
        else:
            from lazyops.utils.system import get_local_kubeconfig
            kconfig = get_local_kubeconfig(name = name, set_as_envval = False)
            if not name: name = pathlib.Path(kconfig).stem
            self.k8s_kubeconfigs[name] = kconfig
        if set_as_envval: os.environ['KUBECONFIG'] = kconfig
        if not self.k8s_active_ctx or set_active: self.k8s_active_ctx = name
        return kconfig

    def bind_settings(self, settings: 'BaseSettings'):
        """
        Binds the settings to this state
        """
        # puts it in here so that we avoid type checking at the module level.
        self.ctx['settings'] = settings
        self.ctx['process_id'] = os.getpid()
        if hasattr(settings, 'app_module_name'):
            self.app_module_name = settings.app_module_name
        else:
            self.app_module_name = settings.__class__.__module__.split(".")[0]
            # settings.app_module_name = settings.__class__.__module__.split(".")[0]
        if hasattr(settings, 'data_path'):
            self.data_path = settings.data_path
        else:
            from lazyops.utils.assets import get_module_path
            module_path = get_module_path(self.app_module_name)
            self.data_path = module_path.joinpath('.data')
        
        self.configure_stx()
    
    def on_exit(self):
        """
        Called on exit
        """
        if 'stx' in self.ctx and self.process_id == self.stx['primary_process_id']:
            logger.info(f"Removing AppState File: {self.ctx['stx_filepath']}", colored = True, prefix = f"|r|State: {self.app_module_name}|e|")
            with contextlib.suppress(FileNotFoundError):
                self.stx.close()
                os.unlink(self.ctx['stx_filepath'])
                os.unlink(self.stx.filelock_path.as_posix())


    def configure_stx(self):
        """
        Configures the stateful statefuldata
        """
        if 'stx' not in self.ctx:
            stx_filepath = self.data_path.joinpath(f'{self.app_module_name}.state.json')
            self.ctx['stx_filepath'] = stx_filepath.as_posix()
            stx = StateData(filepath = stx_filepath)
            self.ctx['stx'] = stx
            atexit.register(self.on_exit)

    def set_primary_process_id(self, process_id: int):
        # sourcery skip: class-extract-method
        """
        Sets the primary process id
        """
        self.stx['primary_process_id'] = process_id
        # logger.info(f"Primary Process ID: {process_id}", colored = True, prefix = "|g|State|e|")
    
    def set_primary_server_process_id(self, process_id: Optional[int] = None):
        """
        Sets the primary server process id
        """
        # Try to find it
        if 'primary_server_process_id' in self.stx.keys(): return
        if process_id is None and self.server_process_id_path.exists():
            with contextlib.suppress(Exception):
                process_id = int(self.server_process_id_path.read_text())
        if process_id is None:
            return
        
        self.stx['primary_server_process_id'] = process_id
        logger.info(f"Primary Server Process ID: {process_id}", colored = True, prefix = "|g|State|e|")
    
    def add_leader_process_id(self, process_id: int, kind: Optional[str] = None):
        """
        Adds a leader process id
        """
        self.stx.append('leader_process_ids', process_id)
        # logger.info(f"Leader Process ID: {process_id} ({kind})", colored = True, prefix = "|g|State|e|")

    def has_logged(self, key: str) -> bool:
        """
        Returns whether the key has been logged
        """
        return self.stx.append('logged', key)
        


from lazyops import lazy_init, get_logger, timed_cache, tstamp
lazy_init('dill')

import dill as pickler
import simdjson as json
import os

from pathlib import Path
from enum import Enum
from abc import ABCMeta
from typing import Union, List, Any, TypeVar, Optional, Dict
from fileio import File, gfile, PathIO, PathIOLike
from tempfile import TemporaryFile, NamedTemporaryFile
from io import BytesIO

logger = get_logger(name='LazyIO')



RESPONSE_SUCCESS_CODES = [
    200, 
    201, 
    202, 
    204
]

VALID_REQUEST_KWARGS = [
    'method',
    'url',
    'content',
    'data',
    'files',
    'json',
    'params',
    'headers',
    'cookies',
    'auth',
    'follow_redirects',
    'timeout',
    'extensions',
]




"""
Import Handler for `psutil`
"""

from lazyops.utils.imports import resolve_missing, require_missing_wrapper

try:
    import psutil
    _psutil_available = True
except ImportError:
    psutil = object
    _psutil_available = False

def resolve_psutil(
    required: bool = False,
):
    """
    Ensures that `psutil` is available
    """
    global psutil, _psutil_available
    if not _psutil_available:
        resolve_missing('psutil', required = required)
        import psutil
        _psutil_available = True
        globals()['psutil'] = psutil


def require_psutil(
    required: bool = False,
):
    """
    Wrapper for `resolve_psutil` that can be used as a decorator
    """
    def decorator(func):
        return require_missing_wrapper(
            resolver = resolve_psutil, 
            func = func, 
            required = required
        )
    return decorator


__all__ = ["GrantOn", "GrantTo", "Privilege"]

from lazyops.libs.dbinit.data_structures.grant_on import GrantOn
from lazyops.libs.dbinit.data_structures.grant_to import GrantTo
from lazyops.libs.dbinit.data_structures.privileges import Privilege




import threading
import contextlib
from typing import Any, Type, Tuple, Dict, List, Union, Optional, Callable, TypeVar, Generic, TYPE_CHECKING


ProxyObjT = TypeVar('ProxyObjT')


class ProxyObject(Generic[ProxyObjT]):
    def __init__(
        self,
        obj_cls: Optional[Union[Type[ProxyObjT], str]] = None,
        obj_getter: Optional[Union[Callable, str]] = None,
        obj_args: Optional[List[Any]] = None,
        obj_kwargs: Optional[Dict[str, Any]] = None,
        obj_initialize: Optional[bool] = True,
        threadsafe: Optional[bool] = True,
        debug_enabled: Optional[bool] = False,
    ):
        """
        Proxy Object

        args:
            obj_cls: the class of the object
            obj_getter: the function to get the object
            debug_enabled: if True, will raise an error if the object is not found
        """
        # Intentionally underscore on the end to avoid conflicts with the settings
        # Evaluate better approach with https://github.com/seperman/dotobject/blob/master/dot/borrowed_lazy.py
        assert obj_cls or obj_getter, "Either `obj_cls` or `obj_getter` must be provided"
        self.__obj_cls_ = obj_cls
        self.__obj_getter_ = obj_getter
        if self.__obj_getter_ and isinstance(self.__obj_getter_, str):
            from lazyops.utils.helpers import lazy_import
            self.__obj_getter_ = lazy_import(self.__obj_getter_)
        self.__threadlock_ = None if threadsafe else threading.Lock()
        self.__obj_ = None
        self.__obj_args_ = obj_args or []
        self.__obj_kwargs_ = obj_kwargs or {}
        self.__obj_initialize_ = obj_initialize
        self.__debug_enabled_ = debug_enabled
        self.__last_attrs_: Dict[str, int] = {}

    @contextlib.contextmanager
    def _objlock_(self):
        """
        Returns the object lock
        """
        if self.__threadlock_ is not None:
            try:
                with self.__threadlock_:
                    yield
            except Exception as e:
                raise e
        else:
            yield

    @property
    def _obj_(self) -> ProxyObjT:
        """
        Returns the object
        """
        if self.__obj_ is None:
            with self._objlock_():
                if self.__obj_getter_:
                    self.__obj_ = self.__obj_getter_(*self.__obj_args_, **self.__obj_kwargs_)
                elif self.__obj_cls_:
                    if isinstance(self.__obj_cls_, str):
                        from lazyops.utils.helpers import lazy_import
                        self.__obj_cls_ = lazy_import(self.__obj_cls_)
                    if self.__obj_initialize_:
                        self.__obj_ = self.__obj_cls_(*self.__obj_args_, **self.__obj_kwargs_)
                    else:
                        self.__obj_ = self.__obj_cls_
        return self.__obj_
        
    def __getattr__(self, name) -> Any:
        """
        Forward all unknown attributes to the proxy object
        """
        if not self.__debug_enabled_:
            return getattr(self._obj_, name)
        
        # Try to debug the attribute
        if name not in self.__last_attrs_:
            self.__last_attrs_[name] = 0
        self.__last_attrs_[name] += 1
        if self.__last_attrs_[name] > 5:
            raise AttributeError(f"Proxy object has no attribute {name}")

        if hasattr(self._obj_, name):
            self.__last_attrs_[name] = 0
            return getattr(self._obj_, name)
        raise AttributeError(f"Settings object has no attribute {name}")
    
    if TYPE_CHECKING:
        def __new__(cls, *args, **kwargs) -> ProxyObjT:
            ...


from .base import BaseSerializer, BinaryBaseSerializer, ObjectValue, SchemaType
from ._json import JsonSerializer
from ._pickle import PickleSerializer
from ._msgpack import MsgPackSerializer
from typing import Any, Dict, Optional, Union, Type

SerializerT = Union[JsonSerializer, PickleSerializer, MsgPackSerializer, BaseSerializer]

def get_serializer(
    serializer_type: Optional[str] = None,
    **kwargs
) -> SerializerT:
    """
    Returns a Serializer
    """
    serializer_type = serializer_type or "pickle"
    if serializer_type in {"json", "orjson", "ujson", "simdjson"}:
        if serializer_type != "json" and "jsonlib" not in kwargs:
            kwargs["jsonlib"] = serializer_type
        return JsonSerializer(**kwargs)
    if serializer_type in {"pickle", "dill", "cloudpickle"}:
        if serializer_type != "pickle" and "picklelib" not in kwargs:
            kwargs["picklelib"] = serializer_type
        return PickleSerializer(**kwargs)
    if serializer_type == "msgpack":
        return MsgPackSerializer(**kwargs)
    raise ValueError(f"Invalid Serializer Type: {serializer_type}")



"""
Import Handler for `bs4`
"""

from lazyops.utils.imports import resolve_missing, require_missing_wrapper

try:
    import bs4
    _bs4_available = True
except ImportError:
    bs4 = object
    _bs4_available = False

def resolve_bs4(
    required: bool = False,
):
    """
    Ensures that `bs4` is available
    """
    global bs4, _bs4_available
    if not _bs4_available:
        resolve_missing('bs4', required = required)
        import bs4
        _bs4_available = True
        globals()['bs4'] = bs4


def require_bs4(
    required: bool = False,
):
    """
    Wrapper for `resolve_bs4` that can be used as a decorator
    """
    def decorator(func):
        return require_missing_wrapper(
            resolver = resolve_bs4, 
            func = func, 
            required = required
        )
    return decorator


import json
import inspect
from lazyops.utils.helpers import is_coro_func
from typing import Callable, Optional, List


def parse_to_list(value: Optional[str] = None) -> Optional[List[str]]:
    """
    Parses a string to a list
    """
    if value is None: return None
    if '[' in value and ']' in value: 
        if '"' in value or "'" in value:
            return json.loads(value)
        value = value.replace('[', '').replace(']', '')
    return next(
        (
            [v.strip() for v in value.split(sep) if v.strip()]
            for sep in [',', ';', '|']
            if sep in value
        ),
        [value],
    )


def create_function_wrapper(function: Callable):
    """
    Creates a function wrapper as a decorator for fastapi
    """
    def inner_wrapper(handler: Callable):
        """
        The inner wrapper
        """
        async def wrapper(*args, **kwargs):
            if is_coro_func(function):
                await function(*args, **kwargs)
            else:
                function(*args, **kwargs)
            return await handler(*args, **kwargs)

        wrapper.__signature__ = inspect.Signature(
            parameters = [
                # Use all parameters from handler
                *inspect.signature(handler).parameters.values(),

                # Skip *args and **kwargs from wrapper parameters:
                *filter(
                    lambda p: p.kind not in (inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD),
                    inspect.signature(wrapper).parameters.values()
                )
            ],
            return_annotation = inspect.signature(handler).return_annotation,
        )

        return wrapper
    return inner_wrapper


def create_post_function_hook_wrapper(function: Callable):
    """
    Creates a function wrapper that executes after the function
    """

    def inner_wrapper(handler: Callable):
        """
        The inner wrapper
        """
        async def wrapper(*args, **kwargs):
            result = await handler(*args, **kwargs)
            if is_coro_func(function): await function(result, *args, **kwargs)
            else: function(result, *args, **kwargs)
            return result
        
        return wrapper
    
    return inner_wrapper

        


from dataclasses import dataclass
from typing import Sequence

from lazyops.libs.dbinit.data_structures.privileges import Privilege
from lazyops.libs.dbinit.mixins.grantable import Grantable


@dataclass
class GrantOn:
    """
    Represents a Sequence of :class:`lazyops.libs.dbinit.data_structures.Privilege` to grant on this
    :class:`lazyops.libs.dbinit.entities.Role` for a Sequence of :class:`lazyops.libs.dbinit.mixins.Grantable`.
    """

    privileges: Sequence[Privilege]
    on: Sequence[Grantable]


# type to represent how to store grants in Role
GrantStore = dict[Grantable, set[Privilege]]



"""
Redis Persistence Type leveraging `aiokeydb` library
"""

from typing import Any, Dict, Optional, Union, Iterable, List, Type, TYPE_CHECKING
from lazyops.utils.lazy import get_keydb_session
from .base import BaseStatefulBackend, SchemaType, logger

if TYPE_CHECKING:
    from aiokeydb import KeyDBSession
    from lazyops.types.models import BaseSettings


class RedisStatefulBackend(BaseStatefulBackend):
    """
    Implements a Redis Stateful Backend
    """
    name: Optional[str] = "redis"

    # Default Global Level Settings that can then be tuned
    # at an instance level
    expiration: Optional[int] = None
    hset_disabled: Optional[bool] = False
    keyjoin: Optional[str] = ':'

    def __init__(
        self,
        name: Optional[str] = None,
        expiration: Optional[int] = None,
        hset_disabled: Optional[bool] = False,
        keyjoin: Optional[str] = None,
        serializer: Optional[str] = None,
        serializer_kwargs: Optional[Dict[str, Any]] = None,
        base_key: Optional[str] = None,
        async_enabled: Optional[bool] = False,
        settings: Optional['BaseSettings'] = None,
        **kwargs,
    ):
        """
        Initializes the backend
        """
        _kdb_kwargs = {}
        if kwargs:
            for key in kwargs:
                if key.startswith('redis_'):
                    _kdb_kwargs[key.replace('redis_', '')] = kwargs.pop(key)
                elif key.startswith('keydb_'):
                    _kdb_kwargs[key.replace('keydb_', '')] = kwargs.pop(key)
        super().__init__(
            serializer=serializer,
            serializer_kwargs=serializer_kwargs,
            base_key=base_key,
            async_enabled=async_enabled,
            settings=settings,
            **kwargs,
        )
        if name is not None: self.name = name
        if expiration is not None: self.expiration = expiration
        self.hset_enabled = (not hset_disabled and self.base_key is not None)
        if keyjoin is not None: self.keyjoin = keyjoin
        self.cache: 'KeyDBSession' = get_keydb_session(
            name = self.name, 
            **_kdb_kwargs
        )

    def get_key(self, key: str) -> str:
        """
        Gets a Key
        """
        if not self.base_key: return key
        return key if self.base_key in key else f'{self.base_key}{self.keyjoin}{key}'
    
    def get(self, key: str, default: Optional[Any] = None, **kwargs) -> Optional[Any]:
        """
        Gets a Value from the DB
        """
        if self.hset_enabled: value = self.cache.hget(self.base_key, key)
        else: value = self.cache.client.get(self.get_key(key))
        if value is None: return default
        try:
            return self.decode_value(value)
        except Exception as e:
            logger.error(f'Error Getting Value for Key: {key} - {e}')
            self.delete(key)
            return default

    def get_values(self, keys: Iterable[str]) -> List[Any]:
        """
        Gets a Value from the DB
        """
        if self.hset_enabled: values = self.cache.hmget(self.base_key, keys)
        else: values = self.cache.client.mget([self.get_key(key) for key in keys])
        results = []
        for key, value in zip(keys, values):
            try:
                results.append(self.decode_value(value))
            except Exception as e:
                logger.error(f'Error Getting Value for Key: {key} - {e}')
                self.delete(key)
                results.append(None)
        return results


    def set(self, key: str, value: Any, ex: Optional[int] = None, **kwargs) -> None:
        """
        Saves a Value to the DB
        """
        ex = ex or self.expiration
        if self.hset_enabled:
            self.cache.hset(self.base_key, key, self.encode_value(value))
            if ex is not None: self.cache.expire(self.base_key, ex)
        else:
            self.cache.client.set(self.get_key(key), self.encode_value(value), ex = ex)
    
    def set_batch(self, data: Dict[str, Any], ex: Optional[int] = None, **kwargs) -> None:
        """
        Saves a Value to the DB
        """
        ex = ex or self.expiration
        data = {k: self.encode_value(v) for k, v in data.items()}
        if self.hset_enabled:
            self.cache.client.hset(self.base_key, mapping = data)
            if ex is not None: self.cache.expire(self.base_key, ex)
        else:
            if self.base_key: data = {self.get_key(k): v for k, v in data.items()}
            self.cache.client.mset(data)
            if ex is not None:
                for key in data:
                    self.cache.client.expire(key, ex)


    def delete(self, key: str, **kwargs) -> None:
        """
        Deletes a Value from the DB
        """
        if self.hset_enabled: self.cache.hdel(self.base_key, key)
        else: self.cache.client.delete(self.get_key(key))

    def clear(self, *keys, **kwargs):
        """
        Clears the Cache
        """
        if self.hset_enabled:
            if keys: self.cache.hdel(self.base_key, *keys)
            else: self.cache.delete(self.base_key)
        elif keys:
            keys = [self.get_key(key) for key in keys]
            self.cache.delete(*keys)
        else:
            keys = self.cache.client.keys(f'{self.base_key}{self.keyjoin}*')
            if keys: self.cache.client.delete(*keys)
    
    async def aget(self, key: str, default: Optional[Any] = None, **kwargs) -> Optional[Any]:
        """
        Gets a Value from the DB
        """
        if self.hset_enabled: value = await self.cache.async_hget(self.base_key, key)
        else: value = await self.cache.async_client.get(self.get_key(key))
        if value is None: return default
        try:
            return self.decode_value(value)
        except Exception as e:
            logger.error(f'Error Getting Value for Key: {key} - {e}')
            await self.adelete(key)
            return default

    async def aget_values(self, keys: Iterable[str]) -> List[Any]:
        """
        Gets a Value from the DB
        """
        if self.hset_enabled: values = await self.cache.async_hmget(self.base_key, keys)
        else: values = await self.cache.async_client.mget([self.get_key(key) for key in keys])
        results = []
        for key, value in zip(keys, values):
            try: results.append(self.decode_value(value))
            except Exception as e:
                logger.error(f'Error Getting Value for Key: {key} - {e}')
                await self.adelete(key)
                results.append(None)
        return results
        
    async def aset(self, key: str, value: Any, ex: Optional[int] = None, **kwargs) -> None:
        """
        Saves a Value to the DB
        """
        ex = ex or self.expiration
        if self.hset_enabled:
            await self.cache.async_hset(self.base_key, key, self.encode_value(value))
            if ex is not None: await self.cache.async_expire(self.base_key, ex)
        else:
            await self.cache.async_client.set(self.get_key(key), self.encode_value(value), ex = ex)

    async def aset_batch(self, data: Dict[str, Any], ex: Optional[int] = None, **kwargs) -> None:
        """
        Saves a Value to the DB
        """
        ex = ex or self.expiration
        data = {k: self.encode_value(v) for k, v in data.items()}
        if self.hset_enabled:
            await self.cache.async_client.hset(self.base_key, mapping = data)
            if ex is not None: await self.cache.async_expire(self.base_key, ex)
        else:
            if self.base_key: data = {self.get_key(k): v for k, v in data.items()}
            await self.cache.async_client.mset(data)
            if ex is not None:
                for key in data:
                    await self.cache.async_client.expire(key, ex)
    
    async def adelete(self, key: str, **kwargs) -> None:
        """
        Deletes a Value from the DB
        """
        if self.hset_enabled: await self.cache.async_hdel(self.base_key, key)
        else: await self.cache.async_client.delete(self.get_key(key))

    async def aclear(self, *keys, **kwargs):
        """
        Clears the Cache
        """
        if self.hset_enabled:
            if keys: await self.cache.async_hdel(self.base_key, *keys)
            else: await self.cache.async_delete(self.base_key)
        elif keys:
            keys = [self.get_key(key) for key in keys]
            await self.cache.async_delete(*keys)

        else:
            keys = await self.cache.async_client.keys(f'{self.base_key}{self.keyjoin}*')
            if keys: await self.cache.async_client.delete(*keys)

    def iterate(self, **kwargs) -> Iterable[Any]:
        """
        Iterates over the Cache
        """
        if self.hset_enabled: return iter(self.cache.hkeys(self.base_key))
        if not self.base_key:
            raise NotImplementedError('Cannot iterate over a Redis Cache without a base key')
        return iter(self.cache.client.keys(f'{self.base_key}{self.keyjoin}*'))
    
    def __len__(self):
        """
        Returns the Length of the Cache
        """
        if self.hset_enabled: return self.cache.hlen(self.base_key)
        if not self.base_key:
            raise NotImplementedError('Cannot get the length of a Redis Cache without a base key')
        return len(self.cache.client.keys(f'{self.base_key}{self.keyjoin}*'))
    

    def get_all_data(self, exclude_base_key: Optional[bool] = False) -> Dict[str, Any]:
        """
        Loads all the Data
        """
        if not self.hset_enabled and not self.base_key:
            raise NotImplementedError('Cannot get all data from a Redis Cache without a base key')
        if self.hset_enabled:
            data = self.cache.hgetall(self.base_key)
            results = {}
            for key, value in data.items():
                if isinstance(key, bytes): key = key.decode()
                try: results[key] = self.decode_value(value)
                except AttributeError:
                    logger.warning(f'Unable to decode value for {key}')
                    self.delete(key)
            return results
        
        keys = self._fetch_set_keys(decode = True)
        data_list = self.cache.client.mget(keys)
        results: Dict[str, Any] = {}
        for key, value in zip(keys, data_list):
            if isinstance(key, bytes): key = key.decode()
            try:
                results[key] = self.decode_value(value)
            except AttributeError:
                logger.warning(f'Unable to decode value for {key}')
                self.delete(key)
        if exclude_base_key:
            results = {k.replace(f'{self.base_key}.', ''): v for k, v in results.items()}
        return results
    
    def get_all_keys(self, exclude_base_key: Optional[bool] = False) -> List[str]:
        """
        Returns all the Keys
        """
        if not self.base_key:
            raise NotImplementedError('Cannot get all keys from a Redis Cache without a base key')
        if self.hset_enabled: return self._fetch_hset_keys(decode = True)
        keys = self._fetch_set_keys(decode = True)
        if exclude_base_key:
            keys = [key.replace(f'{self.base_key}.', '') for key in keys]
        return keys
    
    def get_all_values(self) -> List[Any]:
        """
        Returns all the Values
        """
        if not self.base_key:
            raise NotImplementedError('Cannot get all values from a Redis Cache without a base key')
        if self.hset_enabled:
            data = self.cache.hgetall(self.base_key)
            results = []
            for key, value in data.items():
                try:
                    results.append(self.decode_value(value))
                except Exception as e:
                    logger.warning(f'Unable to decode value for {key}: {e}')
                    self.delete(key)
            return results
        keys = self._fetch_set_keys(decode = False)
        data_list = self.cache.client.mget(keys)
        results = []
        for key, value in zip(keys, data_list):
            try:
                results.append(self.decode_value(value))
            except Exception as e:
                logger.warning(f'Unable to decode value for {key}: {e}')
                self.delete(key)
        return results

    async def aget_all_data(self, exclude_base_key: Optional[bool] = False) -> Dict[str, Any]:
        """
        Loads all the Data
        """
        if not self.base_key:
            raise NotImplementedError('Cannot get all data from a Redis Cache without a base key')
        if self.hset_enabled:
            data = await self.cache.async_hgetall(self.base_key)
            results = {}
            for key, value in data.items():
                if isinstance(key, bytes): key = key.decode()
                try:
                    results[key] = self.decode_value(value)
                except Exception as e:
                    logger.warning(f'Unable to decode value for {key}: {e}')
                    await self.adelete(key)
            return results
        keys = await self._afetch_set_keys(decode = True)
        data_list = await self.cache.async_client.mget(keys)
        results: Dict[str, Any] = {}
        for key, value in zip(keys, data_list):
            try:
                results[key] = self.decode_value(value)
            except Exception as e:
                logger.warning(f'Unable to decode value for {key}: {e}')
                await self.adelete(key)
        if exclude_base_key:
            results = {k.replace(f'{self.base_key}.', ''): v for k, v in results.items()}
        return results
    
    async def aget_all_keys(self, exclude_base_key: Optional[bool] = False) -> List[str]:
        """
        Returns all the Keys
        """
        if not self.base_key:
            raise NotImplementedError('Cannot get all keys from a Redis Cache without a base key')
        if self.hset_enabled: 
            return await self._afetch_hset_keys(decode = True)
        keys = await self._afetch_set_keys(decode = True)
        if exclude_base_key:
            keys = [key.replace(f'{self.base_key}.', '') for key in keys]
        return keys
    
    async def aget_all_values(self) -> List[Any]:
        """
        Returns all the Values
        """
        if not self.base_key:
            raise NotImplementedError('Cannot get all values from a Redis Cache without a base key')
        if self.hset_enabled:
            data = await self.cache.async_hgetall(self.base_key)
            results = []
            for key, value in data.items():
                try:
                    results.append(self.decode_value(value))
                except Exception as e:
                    logger.warning(f'Unable to decode value for {key}: {e}')
                    await self.adelete(key)
            return results
        keys = await self._afetch_set_keys(decode = False)
        data_list = await self.cache.async_client.mget(keys)
        results = []
        for key, value in zip(keys, data_list):
            try:
                results.append(self.decode_value(value))
            except Exception as e:
                logger.warning(f'Unable to decode value for {key}: {e}')
                await self.adelete(key)
        return results

    def contains(self, key):
        """
        Returns True if the Cache contains the Key
        """
        if self.hset_enabled: return self.cache.hexists(self.base_key, key)
        return self.cache.client.exists(self.get_key(key))
    
    async def acontains(self, key):
        """
        Returns True if the Cache contains the Key
        """
        if self.hset_enabled: return await self.cache.async_hexists(self.base_key, key)
        return await self.cache.async_client.exists(self.get_key(key))
    
    def expire(self, key: str, ex: int) -> None:
        """
        Expires a Key
        """
        if self.hset_enabled: 
            logger.warning(f'Cannot expire a key in a hset cache: {self.base_key}.{key}')
            return
        self.cache.client.expire(self.get_key(key), ex)

    async def aexpire(self, key: str, ex: int) -> None:
        """
        Expires a Key
        """
        if self.hset_enabled: 
            logger.warning(f'Cannot expire a key in a hset cache: {self.base_key}.{key}')
            return
            # await self.cache.async_expire(self.base_key, ex)
        await self.cache.async_client.expire(self.get_key(key), ex)


    """
    Utility Functions
    """

    def _fetch_set_keys(self, decode: Optional[bool] = True) -> List[str]:
        """
        This is a utility func for non-hset
        """
        keys: List[Union[str, bytes]] = self.cache.client.keys(f'{self.base_key}{self.keyjoin}*')
        if decode: return [key.decode() if isinstance(key, bytes) else key for key in keys]
        return keys
    
    def _fetch_hset_keys(self, decode: Optional[bool] = True) -> List[str]:
        """
        This is a utility func for hset
        """
        keys: List[Union[str, bytes]] = self.cache.hkeys(self.base_key)
        if decode: return [key.decode() if isinstance(key, bytes) else key for key in keys]
        return keys

    async def _afetch_set_keys(self, decode: Optional[bool] = True) -> List[str]:
        """
        This is a utility func for non-hset
        """
        keys: List[Union[str, bytes]] = await self.cache.async_client.keys(f'{self.base_key}{self.keyjoin}*')
        if decode: return [key.decode() if isinstance(key, bytes) else key for key in keys]
        return keys
    
    async def _afetch_hset_keys(self, decode: Optional[bool] = True) -> List[str]:
        """
        This is a utility func for hset
        """
        keys: List[Union[str, bytes]] = await self.cache.async_hkeys(self.base_key)
        if decode: return [key.decode() if isinstance(key, bytes) else key for key in keys]
        return keys
    

    def migrate_compression(self, **kwargs) -> List[str]:
        """
        Migrates the Compression
        """
        if not self.hset_enabled and not self.base_key:
            raise NotImplementedError('Cannot get all data from a Redis Cache without a base key')
        failed_keys = []
        null_is_error = self.serializer.raise_errors
        if self.hset_enabled:
            data = self.cache.hgetall(self.base_key)
            results = {}
            for key, value in data.items():
                if isinstance(key, bytes): key = key.decode()
                try: 
                    val = self.decode_value(value)
                    if val is not None: results[key] = val
                    elif null_is_error: failed_keys.append(key)
                except Exception as e:
                    if not null_is_error:
                        logger.warning(f'Unable to decode value for {key}: {e}')
                    failed_keys.append(key)
                    # self.delete(key)
            if results: self.set_batch(results)
            if failed_keys: logger.warning(f'Failed to migrate keys: {failed_keys}')
            return failed_keys
        
        keys = self._fetch_set_keys(decode = True)
        data_list = self.cache.client.mget(keys)
        results: Dict[str, Any] = {}
        for key, value in zip(keys, data_list):
            if isinstance(key, bytes): key = key.decode()
            try:
                val = self.decode_value(value)
                if val is not None: results[key] = val
                elif null_is_error: failed_keys.append(key)
            except Exception as e:
                if not null_is_error:
                    logger.warning(f'Unable to decode value for {key}: {e}')
                failed_keys.append(key)
        
        if results: self.set_batch(results)
        if failed_keys: logger.warning(f'Failed to migrate keys: {failed_keys}')
        return failed_keys
        

    async def amigrate_compression(self, **kwargs) -> List[str]:
        """
        Migrates the Compression
        """
        if not self.hset_enabled and not self.base_key:
            raise NotImplementedError('Cannot get all data from a Redis Cache without a base key')
        failed_keys = []
        null_is_error = self.serializer.raise_errors
        if self.hset_enabled:
            data = await self.cache.async_hgetall(self.base_key)
            results = {}
            for key, value in data.items():
                if isinstance(key, bytes): key = key.decode()
                try: 
                    val = await self.adecode_value(value)
                    if val is not None: results[key] = val
                    elif null_is_error: failed_keys.append(key)
                except Exception as e:
                    if not null_is_error:
                        logger.warning(f'Unable to decode value for {key}: {e}')
                    failed_keys.append(key)
            if results: await self.aset_batch(results)
            if failed_keys: logger.warning(f'Failed to migrate keys: {failed_keys}')
            return failed_keys
        
        keys = await self._afetch_set_keys(decode = True)
        data_list = await self.cache.async_client.mget(keys)
        results: Dict[str, Any] = {}
        for key, value in zip(keys, data_list):
            if isinstance(key, bytes): key = key.decode()
            try:
                val = await self.adecode_value(value)
                if val is not None: results[key] = val
                elif null_is_error: failed_keys.append(key)
            except Exception as e:
                if not null_is_error:
                    logger.warning(f'Unable to decode value for {key}: {e}')
                failed_keys.append(key)
        
        if results: await self.aset_batch(results)
        if failed_keys: logger.warning(f'Failed to migrate keys: {failed_keys}')
        return failed_keys


"""
The Base Configuration Schema
"""

import os
import functools
from enum import Enum
from pathlib import Path

from lazyops.types.models import BaseSettings, pre_root_validator, validator
from lazyops.imports._pydantic import BaseAppSettings, BaseModel
from lazyops.utils.system import is_in_kubernetes, get_host_name
from lazyops.utils.assets import create_get_assets_wrapper, create_import_assets_wrapper
from lazyops.libs.fastapi_utils import GlobalContext
from lazyops.libs.fastapi_utils.types.persistence import TemporaryData
from typing import List, Optional, Dict, Any, Callable, Union, Type, TYPE_CHECKING


_temp_datas: Dict[str, TemporaryData] = {}
_get_assets_wrappers: Dict[str, Callable[..., Union[Dict[str, Any], List[Any]]]] = {}
_import_assets_wrappers: Dict[str, Callable[..., Union[Path, Any, List[Path], List[Any], Dict[str, Path], Dict[str, Any]]]] = {}

def get_temp_data(name: str) -> TemporaryData:
    """
    Returns a temporary data object
    """
    global _temp_datas
    if name not in _temp_datas:
        _temp_datas[name] = TemporaryData.from_module(name)
    return _temp_datas[name]

def get_assets_func(module_name: str, asset_path: Optional[str] = "assets") -> Callable[..., Union[Dict[str, Any], List[Any]]]:
    """
    Returns the get assets function
    """
    global _get_assets_wrappers
    if module_name not in _get_assets_wrappers:
        _get_assets_wrappers[module_name] = create_get_assets_wrapper(module_name, asset_path)
    return _get_assets_wrappers[module_name]

def import_assets_func(module_name: str, asset_path: Optional[str] = "assets") -> Callable[..., Union[Path, Any, List[Path], List[Any], Dict[str, Path], Dict[str, Any]]]:
    """
    Returns the import assets function
    """
    global _import_assets_wrappers
    if module_name not in _import_assets_wrappers:
        _import_assets_wrappers[module_name] = create_import_assets_wrapper(module_name, asset_path)
    return _import_assets_wrappers[module_name]


class AppEnv(str, Enum):
    CICD = "cicd"
    DEVELOPMENT = "development"
    STAGING = "staging"
    PRODUCTION = "production"
    LOCAL = "local"

    @classmethod
    def from_env(cls, env_value: str) -> "AppEnv":
        """
        Get the app environment from the environment variables
        """
        env_value = env_value.lower()
        if "cicd" in env_value or "ci/cd" in env_value:
            return cls.CICD
        if "prod" in env_value:
            return cls.PRODUCTION
        if "dev" in env_value:
            return cls.DEVELOPMENT
        if "staging" in env_value:
            return cls.STAGING
        if "local" in env_value:
            return cls.LOCAL
        raise ValueError(f"Invalid app environment: {env_value} ({type(env_value)})")
    
    def __eq__(self, other: Any) -> bool:
        """
        Equality operator
        """
        if isinstance(other, str):
            return self.value == other.lower()
        return self.value == other.value if isinstance(other, AppEnv) else False

    @property
    def is_devel(self) -> bool:
        """
        Returns True if the app environment is development
        """
        return self in [
            AppEnv.LOCAL,
            AppEnv.CICD,
            AppEnv.DEVELOPMENT
        ]

    @property
    def is_local(self) -> bool:
        """
        Returns True if the app environment is local
        """
        return self in [
            AppEnv.LOCAL,
            AppEnv.CICD,
        ]

    @property
    def name(self) -> str:
        """
        Returns the name in lower
        """
        return self.value.lower()


def get_app_env(name: Optional[str] = None) -> AppEnv:
    """
    Retrieves the app environment
    """
    for key in {
        "SERVER_ENV",
        f"{name.upper()}_ENV" if name is not None else "APP_SERVER_ENV",
        "APP_ENV",
        "ENVIRONMENT",
    }:
        if env_value := os.getenv(key):
            return AppEnv.from_env(env_value)

    if is_in_kubernetes():
        parts = get_host_name().split("-")
        return AppEnv.from_env(parts[2]) if len(parts) > 3 else AppEnv.PRODUCTION
    return AppEnv.LOCAL

def get_app_ingress(
    module_name: str,
    base_domain: str,
    app_env: AppEnv,
    app_host: Optional[str] = None,
    app_port: Optional[int] = None,
    
) -> str:
    """
    Retrieves the app ingress
    """
    if app_env.is_local:
        app_host = app_host or "localhost"
        app_port = app_port or 8080
        return f"http://{app_host}:{app_port}"
    if app_env == AppEnv.DEVELOPMENT:
        return f"https://{module_name}-develop.{base_domain}"
    if app_env == AppEnv.STAGING:
        return f"https://{module_name}-staging.{base_domain}"
    if app_env == AppEnv.PRODUCTION:
        return f"https://{module_name}.{base_domain}"
    raise ValueError(f"Invalid app environment: {app_env}")



def get_app_env_file(
    configs_path: Path,
    name: Optional[str] = None, 
    required: Optional[bool] = False,
    module_name: Optional[str] = None,
) -> Optional[Path]:
    """
    Retrieves the app environment file

    Only valid for local/dev environments
    """
    app_env = get_app_env(module_name)
    is_local_env = app_env in [
        AppEnv.LOCAL,
        AppEnv.DEVELOPMENT,
        AppEnv.CICD,
    ]
    envs_path = configs_path.joinpath('envs')
    if name is not None:
        if envs_path.joinpath(f'{name}-{app_env.name}.env').exists():
            return envs_path.joinpath(f'{name}-{app_env.name}.env')
        if envs_path.joinpath(f'{name}.env').exists():
            return envs_path.joinpath(f'{name}.env')
        if required and not is_local_env: raise ValueError(f"Invalid app environment file: {name}")
    env_path = envs_path.joinpath(f'{app_env.name}.env')
    if env_path.exists(): return env_path
    if is_local_env or not required:
        return envs_path.joinpath('default.env')
    return None

def get_app_default_file(
    configs_path: Path,
    name: Optional[str] = None, 
    required: Optional[bool] = False, 
    suffix: Optional[str] = None,
    module_name: Optional[str] = None,
) -> Optional[Path]:
    """
    Retrieves the app environment file

    Only valid for local/dev environments
    """
    app_env = get_app_env(module_name)
    defaults_path = configs_path.joinpath('defaults')
    suffix = suffix or 'json'
    if name is not None:
        if defaults_path.joinpath(f'{name}-{app_env.name}.{suffix}').exists():
            return defaults_path.joinpath(f'{name}-{app_env.name}.{suffix}')
        if defaults_path.joinpath(f'{name}.{suffix}').exists():
            return defaults_path.joinpath(f'{name}.{suffix}')
        if required: raise ValueError(f"Invalid app environment file: {name}")
    env_path = defaults_path.joinpath(f'{app_env.name}.{suffix}')
    if env_path.exists(): return env_path
    default_path = defaults_path.joinpath(f'default.{suffix}')
    return default_path if default_path.exists() else None

def create_get_app_env_file(
    configs_path: Path,
    module_name: Optional[str] = None,
) -> Callable[[Optional[str], Optional[bool]], Optional[Path]]:
    """
    Creates a get_app_env_file wrapper
    """
    kwargs = {}
    if module_name is not None: kwargs["module_name"] = module_name
    return functools.partial(get_app_env_file, configs_path, **kwargs)

def create_get_app_default_file(
    configs_path: Path,
    module_name: Optional[str] = None,
) -> Callable[[Optional[str], Optional[bool], Optional[str]], Optional[Path]]:
    """
    Creates a get_app_default_file wrapper
    """
    kwargs = {}
    if module_name is not None: kwargs["module_name"] = module_name
    return functools.partial(get_app_default_file, configs_path, **kwargs)

def create_get_app_ingress(
    module_name: str,
    base_domain: str,
) -> Callable[[], str]:
    """
    Creates a get_app_ingress wrapper
    """
    return functools.partial(
        get_app_ingress,
        module_name,
        base_domain,
    )

def create_get_app_env(
    name: str,
) -> Callable[[], AppEnv]:
    """
    Creates a get_app_env wrapper
    """
    return functools.partial(
        get_app_env,
        name,
    )


class AppSettings(BaseAppSettings):
    """
    Custom App Settings
    """

    app_env: Optional[AppEnv] = None 

    @validator('app_env', pre=True)
    def validate_app_env(cls, value: Optional[Any]) -> Any:
        """
        Validates the app environment
        """
        if value is None:
            return get_app_env(cls.__module__.split(".")[0])
        return AppEnv.from_env(value) if isinstance(value, str) else value

    def get_assets(
        self, 
        *path_parts,
        load_file: Optional[bool] = False,
        loader: Optional[Callable] = None,
        **kwargs
    ) -> Union[Path, Any, List[Path], List[Any], Dict[str, Path], Dict[str, Any]]:
        """
        Retrieves the assets

        args:
            path_parts: path parts to the assets directory (default: [])
            load_file: load the file (default: False)
            loader: loader function to use (default: None)
            **kwargs: additional arguments to pass to `get_module_assets`
        """
        get_assets = get_assets_func(self.module_name, "assets")
        return get_assets(*path_parts, load_file = load_file, loader = loader, **kwargs)
    
    def import_assets(
        self, 
        *path_parts,
        model: Optional[Type['BaseModel']] = None,
        load_file: Optional[bool] = False,
        **kwargs
    ) -> Union[Dict[str, Any], List[Any]]:
        """
        Import assets from a module.

        args:
            path_parts: path parts to the assets directory (default: [])
            model: model to parse the assets with (default: None)
            load_file: load the file (default: False)
            **kwargs: additional arguments to pass to import_module_assets
        """
        import_assets = import_assets_func(self.module_name, "assets")
        return import_assets(*path_parts, model = model, load_file = load_file, **kwargs)


    @property
    def is_local_env(self) -> bool:
        """
        Returns whether the environment is development
        """
        return self.app_env in [AppEnv.DEVELOPMENT, AppEnv.LOCAL] and not self.in_k8s
    
    @property
    def is_production_env(self) -> bool:
        """
        Returns whether the environment is production
        """
        return self.app_env == AppEnv.PRODUCTION and self.in_k8s

    @property
    def is_development_env(self) -> bool:
        """
        Returns whether the environment is development
        """
        return self.app_env in [AppEnv.DEVELOPMENT, AppEnv.LOCAL, AppEnv.CICD]



"""
Import Handler for `pycryptodome`
"""

from lazyops.utils.imports import resolve_missing, require_missing_wrapper

try:
    import Crypto
    _pycryptodome_available = True
except ImportError:
    Crypto = object
    _pycryptodome_available = False

def resolve_pycryptodome(
    required: bool = False,
):
    """
    Ensures that `pycryptodome` is available
    """
    global Crypto, _pycryptodome_available
    if not _pycryptodome_available:
        resolve_missing('pycryptodome', required = required)
        import Crypto
        _pycryptodome_available = True
        globals()['Crypto'] = Crypto


def require_pycryptodome(
    required: bool = False,
):
    """
    Wrapper for `resolve_pycryptodome` that can be used as a decorator
    """
    def decorator(func):
        return require_missing_wrapper(
            resolver = resolve_pycryptodome, 
            func = func, 
            required = required
        )
    return decorator


"""
Local Persistence Types
"""

import os
import json
import filelock
import binascii
from pathlib import Path, PurePath
from typing import TypeVar, Generic, Any, Dict, Optional, Union, Iterable, List, Type, TYPE_CHECKING
from lazyops.utils.helpers import  create_unique_id
from .base import BaseStatefulBackend, SchemaType, logger

try:
    from fileio import File, FileLike
except ImportError:
    File = Path
    FileLike = PurePath


if TYPE_CHECKING:
    from lazyops.types.models import BaseSettings


class LocalStatefulBackend(BaseStatefulBackend):
    """
    Implements a Local Stateful Backend
    """
    name: Optional[str] = "local"
    encoding: Optional[str] = "utf-8"

    def __init__(
        self,
        name: Optional[str] = None,
        file_path: Optional[File] = None,
        encoding: Optional[str] = None,

        serializer: Optional[str] = None,
        serializer_kwargs: Optional[Dict[str, Any]] = None,
        base_key: Optional[str] = None,
        async_enabled: Optional[bool] = False,
        settings: Optional['BaseSettings'] = None,
        **kwargs,
    ):
        """
        Initializes the backend
        """
        super().__init__(
            serializer=serializer,
            serializer_kwargs=serializer_kwargs,
            base_key=base_key,
            async_enabled=async_enabled,
            settings=settings,
            **kwargs,
        )
        if name is not None: self.name = name
        if encoding is not None: self.encoding = encoding
        if file_path is None:
            if hasattr(self.settings, "data_dir"):
                file_path = self.settings.data_dir.joinpath(f"{self.name}.cache")
            else:
                file_path = Path(os.getcwd()).joinpath(f"{self.name}.cache")
        self.file_path = File(file_path)
        if self.file_path.is_dir():
            self.file_path = self.file_path.joinpath(f"{self.name}.cache")
        self.file_path.parent.mkdir(parents = True, exist_ok = True)
        self.file_lock_path = self.file_path.with_suffix(".lock")
        self.file_hash_path = self.file_path.with_suffix(".hash")
        self.file_lock = filelock.FileLock(lock_file = self.file_lock_path.as_posix(), thread_local = False)
        try:
            import simdjson
            self.parser = simdjson.Parser()
            self.cache: simdjson.SimValue = {}
        except ImportError:
            self.parser = None
            self.cache: Dict[str, Any] = {}

        if not self.file_path.exists():
            with self.file_lock:
                self.file_path.write_text('{}')
                self.file_hash = create_unique_id()
                self.file_hash_path.write_text(self.file_hash)
        else:
            self.cache = self.get_data()
            self.file_hash = self.file_hash_path.read_text()
    
    def encode_value(self, value: Union[Any, SchemaType], **kwargs) -> str:
        """
        Encodes a Value
        """
        result = super().encode_value(value, **kwargs)
        return result.hex() if isinstance(result, bytes) else result
    
    def decode_value(self, value: str, **kwargs) -> Any:
        """
        Decodes a Value
        """
        if self.serializer.binary or self.serializer.compression_enabled:
            value = binascii.unhexlify(value)
        return super().decode_value(value, **kwargs)
    
    def _precheck(self, **kwargs):
        """
        Precheck
        """
        self.sync()

    def get(self, key: str, default: Optional[Any] = None, **kwargs) -> Optional[Any]:
        """
        Gets a Value from the JSON
        """
        self.sync()
        if value := self.cache.get(self.get_key(key)):
            try:
                result = self.decode_value(value, **kwargs)
                return result if result is not None else default
            except Exception as e:
                logger.info(f'Error Decoding Value: |r|({type(value)}) {e}|e| {value}', colored = True)
                self.delete(key)
                return default
        return default
    
    def get_values(self, keys: Iterable[str]) -> List[Any]:
        """
        Gets a Value from the JSON
        """
        self.sync()
        results = []
        for key in keys:
            if value := self.cache.get(self.get_key(key)):
                try:
                    results.append(self.decode_value(value))
                except AttributeError:
                    logger.warning(f'Unable to decode value for {key}')
                    self.delete(key)
                    results.append(None)
            else:
                results.append(None)
        return results
    

    def set(self, key: str, value: Any, *args, **kwargs) -> None:
        """
        Sets a Value in the JSON
        """
        self.sync()
        self.cache[self.get_key(key)] = self.encode_value(value)
        self.write_data(self.cache)
        

    def set_batch(self, data: Dict[str, Any], *args, **kwargs) -> None:
        """
        Sets a Value in the JSON
        """
        self.sync()
        for key, value in data.items():
            self.cache[self.get_key(key)] = self.encode_value(value)
        self.write_data(self.cache)

    def delete(self, key: str, **kwargs) -> None:
        """
        Deletes a Value from the JSON
        """
        self.sync()
        _ = self.cache.pop(self.get_key(key), None)
        self.write_data(self.cache)
    
    def clear(self, *keys: str, **kwargs):
        """
        Clears the Cache
        """
        self.sync()
        if keys:
            for key in keys:
                self.cache.pop(self.get_key(key), None)
        else:
            self.cache = {}
        self.write_data(self.cache)


    def get_all_keys(self, exclude_base_key: Optional[bool] = False) -> List[str]:
        """
        Returns all the Keys
        """
        self.sync()
        keys = [
            key
            for key in self.cache.keys()
            if key.startswith(self.base_key)
        ] if self.base_key else list(self.cache.keys())
        if exclude_base_key:
            keys = [key.replace(f'{self.base_key}.', '') for key in keys]
        return keys

    def get_all_data(self, exclude_base_key: Optional[bool] = False) -> Dict[str, Any]:
        """
        Loads all the Data
        """
        self.sync()
        all_keys = self.get_all_keys()
        data = {
            key: self.decode_value(value)
            for key, value in self.cache.items() if key in all_keys
        }
        if exclude_base_key:
            data = {key.replace(f'{self.base_key}.', ''): value for key, value in data.items()}
        return data
    
    def get_all_values(self) -> Iterable[Any]:
        """
        Returns all the Values
        """
        self.sync()
        all_keys = self.get_all_keys()
        return [
            self.decode_value(value)
            for key, value in self.cache.items() if key in all_keys
        ]    

    def contains(self, key: str) -> bool:
        """
        Returns True if the Cache contains the Key
        """
        self.sync()
        return key in self.get_all_keys()
    

    def iterate(self, **kwargs) -> Iterable[Any]:
        """
        Iterates over the Cache
        """
        self.sync()
        return iter(self.get_all_keys())


    """
    Primary Utility Methods
    """

    def should_sync(self) -> bool:
        """
        Returns True if the File should be Synced
        """
        return self.file_hash != self.file_hash_path.read_text()
    
    async def ashould_sync(self) -> bool:
        """
        Returns True if the File should be Synced
        """
        return self.file_hash != await self.file_hash_path.async_read_text()
    
    def sync(self) -> None:
        """
        Syncs the File
        """
        if self.should_sync():
            self.cache = self.get_data()

    async def async_sync(self) -> None:
        """
        Syncs the File
        """
        if await self.ashould_sync():
            self.cache = await self.aget_data()

    def get_data(self) -> Dict[str, Any]:
        """
        Returns the Data
        """
        with self.file_lock:
            if self.parser is None:
                return json.loads(self.file_path.read_text())
            return self.parser.parse(self.file_path.read_bytes(), recursive=True)
    
    async def aget_data(self) -> Dict[str, Any]:
        """
        Fetch the data
        """
        with self.file_lock:
            if self.parser is None:
                return json.loads(await self.file_path.async_read_text())
            return self.parser.parse(await self.file_path.async_read_bytes(), recursive=True)
        
    def write_data(self, data: Dict):
        """
        Writes the Data to the File
        """
        with self.file_lock:
            self.file_path.write_text(json.dumps(data, indent = 4, ensure_ascii = False))
            self.file_hash = create_unique_id()
            self.file_hash_path.write_text(self.file_hash)
    
    async def awrite_data(self, data: Dict):
        """
        Writes the Data to the File
        """
        with self.file_lock:
            await self.file_path.async_write_text(json.dumps(data, indent = 4, ensure_ascii = False))
            self.file_hash = create_unique_id()
            await self.file_hash_path.async_write_text(self.file_hash)


from __future__ import annotations

from abc import ABC, abstractmethod
from inspect import signature
from typing import Any, Sequence

from lazyops.libs.dbinit.base import Engine
from lazyops.libs.dbinit.exceptions import EntityExistsError, NoEngineError


class Entity(ABC):
    """
    Base class to inherit from for all entities. Defines methods common to all entities, abstract methods that are
    needed for all entities, and class methods that enable collection of entities for easy cluster interaction.
    """

    entities: list["Entity"] = []
    check_if_any_exist: bool = False
    _engine: Engine | None = None

    def __init__(
        self,
        name: str,
        depends_on: Sequence["Entity"] | None = None,
        check_if_exists: bool | None = None,
    ):
        """
        :param name: Unique name of the entity. Cluster-level entities must be unique, database-level entities must be unique within a database.
        :param depends_on: Any entities that should be created before this one.
        :param check_if_exists: Flag to set existence check behavior. If `True`, will raise an exception during _safe_create if the entity already exists, and will raise an exception during _safe_drop if the entity does not exist.
        """
        # TODO have "name" be a str class that validates via regex for valid postgres names
        self.name = name

        # explicit None check because False requires different behavior
        if check_if_exists is None:
            self.check_if_exists = self.__class__.check_if_any_exist
        else:
            self.check_if_exists = check_if_exists

        if not depends_on:
            depends_on = []
        self.depends_on: Sequence["Entity"] = depends_on

        self.__class__._register(self)
    
    @property
    def encoded_name(self) -> str:
        """
        Returns the encoded name of the entity.
        """
        return f'"{self.name}"' if '-' in self.name or '_' in self.name else self.name

    def __hash__(self) -> int:
        return hash((self.name, self.__class__.__name__))

    def __eq__(self, other: object) -> bool:
        return (
            (self.name, self.__class__.__name__)
            == (other.name, other.__class__.__name__)
            if isinstance(other, self.__class__)
            else NotImplemented
        )

    @classmethod
    def _register(cls, entity: "Entity") -> None:
        """
        Appends an entity to the running list of all entities.
        :param entity: Any given entity, can be thought of as operating on a subclassed `self`.
        """
        # TODO add code to make sure there are no duplicate entities
        cls.entities.append(entity)

    @classmethod
    def engine(cls) -> Engine:
        """
        Convenience wrapper to safely get the main engine.
        :return: A :class:`sqlalchemy.Engine` if one has been assigned, otherwise raises an exception.
        """
        if cls._engine:
            return cls._engine
        else:
            raise NoEngineError("There is no SQLAlchemy Engine present. `Base._engine` must have " "a valid engine.")

    @abstractmethod
    def _create(self) -> None:
        """
        Create this entity in the cluster.
        """
        pass

    def _safe_create(self) -> None:
        """
        Run an existence check before attempting to create the entity in the cluster.
        """
        if not self._exists():
            self._create()
        elif self.check_if_exists or self.check_if_any_exist:
            raise EntityExistsError(
                f"There is already a {self.__class__.__name__} with the "
                f"name {self.name}. If you want to proceed anyway, set "
                f"the `check_if_exists` parameter to False. This will "
                f"simply skip over the existing entity."
            )

    @abstractmethod
    def _exists(self) -> bool:
        """
        Check if this entity currently exists in the cluster.
        :return: True if it exists, False if it does not.
        """
        pass

    @abstractmethod
    def _drop(self) -> None:
        """
        Drop this entity from the cluster.
        """
        pass

    def _safe_drop(self) -> None:
        """
        Run an existence check before attempting to drop the entity from the cluster.
        """
        if self._exists():
            self._drop()
        elif self.check_if_exists or self.check_if_any_exist:
            raise EntityExistsError(
                f"There is no {self.__class__.__name__} with the "
                f"name {self.name} to remove. If you want to proceed "
                f"anyway, set the `check_if_exists` parameter to False. "
                f"This will simply skip over the removal of this "
                f"entity that does not exist in the cluster."
            )

    def _get_passed_args(self) -> dict[str, Any]:
        """
        Helper to grab all the arguments to __init__ that aren't in the superclass and have a non-None value. Useful
        for subclasses.
        :return: A dict mapping the names of init arguments to their values.
        """
        return {
            k: v
            for k, v in vars(self).items()
            if (k not in signature(self.__class__.__bases__[0].__init__).parameters)  # type: ignore
            and (v is not None)
            and (k != "_grant_name")
        }



"""
Server side tasks

Usage:

@register_server_task
async def task1():
    ...

@app.on_event("startup")
async def startup():
    spawn_server_tasks()

"""
import time
import asyncio
import threading

from lazyops.utils.logs import logger
from typing import List, Union, Set, Callable, Tuple, Optional

_server_tasks: Set[asyncio.Task] = set()
_server_task_index: List[Callable] = []
_server_primary_task_index: List[Callable] = []


def register_server_task(
    func: Callable, 
    name: Optional[str] = None,
    primary_process_only: Optional[bool] = None,
):
    """
    Register a server task
    """
    global _server_task_index, _server_primary_task_index
    from lazyops.libs.fastapi_utils.processes import GlobalContext
    if GlobalContext.is_leader_process:
        logger.info(f"Registered server task: {name or func.__name__} [Primary: {primary_process_only}]")
    if primary_process_only:
        _server_primary_task_index.append(func)
    else:
        _server_task_index.append(func)
    return func


def create_server_task(
    name: Optional[str] = None,
    primary_process_only: Optional[bool] = None,
    interval: Optional[float] = 60.0,
    stop_on_error: Optional[bool] = True,
):
    """
    Create a server task and register it with this wrapper
    """
    def decorator(func):
        """
        Inner wrapper
        """
        async def wrapper_func(*args, **kwargs):
            iterations = 0
            while True:
                try:
                    await func(*args, iterations = iterations, **kwargs)
                    iterations += 1
                except Exception as e:
                    logger.trace(f"Error in server task {name or func.__name__}", e)
                    if stop_on_error:
                        logger.info(f"Stopping server task {name or func.__name__}...")
                        break
                await asyncio.sleep(interval)
            
        register_server_task(
            func = wrapper_func, 
            name = name,
            primary_process_only = primary_process_only,
        )
        return wrapper_func
     
    return decorator


async def start_bg_tasks(
    primary_server_process_only: Optional[bool] = None,
):
    """
    Start background tasks
    """
    from lazyops.libs.fastapi_utils.processes import GlobalContext
    
    if primary_server_process_only and not GlobalContext.is_primary_server_process:
        return
    
    for task in _server_task_index:
        task = asyncio.create_task(task())
        _server_tasks.add(task)
        task.add_done_callback(_server_tasks.discard)
    
    if GlobalContext.is_primary_server_process:
        for task in _server_primary_task_index:
            task = asyncio.create_task(task())
            _server_tasks.add(task)
            task.add_done_callback(_server_tasks.discard)

def spawn_server_tasks():
    """
    Spawn server tasks
    """
    proc = threading.Thread(target = start_bg_tasks)
    proc.start()
    while True:
        try:
            time.sleep(5.0)
        except Exception as e:
            logger.info("Stopping server tasks...")
            for task in _server_tasks:
                task.cancel()
            break
    proc.join()



from __future__ import annotations

import kopf
import copy
import datetime
import aiohttp

from kopf._cogs.clients import api, errors
from kopf._cogs.configs import configuration
from kopf._cogs.helpers import typedefs
from kopf._cogs.structs import bodies, references
import kopf._cogs.clients.events
from typing import Dict, Any, TYPE_CHECKING
from lazyops.utils.logs import default_logger as _logger

if TYPE_CHECKING:
    from lazyops.libs.kops.config import KOpsSettings

# Patch so that event logs will show `kops` instead of `kopf` as the source.
_kops_config: Dict[str, Any] = None
_kops_settings: 'KOpsSettings' = None


def _get_config():
    global _kops_config, _kops_settings
    if _kops_settings is None:
        from lazyops.libs.kops.client import KOpsClient
        _kops_settings = KOpsClient.settings
    
    if _kops_config is None:
        _kops_config = {
            'MAX_MESSAGE_LENGTH': _kops_settings.kops_max_message_length,
            'CUT_MESSAGE_INFIX': _kops_settings.kops_cut_message_infix,
            'FINALIZER': _kops_settings.kops_finalizer,
            'PREFIX': _kops_settings.kops_prefix,
            'PERSISTENT_KEY': _kops_settings.kops_persistent_key,
            'GENERATE_EVENT_NAME': _kops_settings.kops_generate_event_name,
            'WATCH_INTERVAL': _kops_settings.kops_watch_interval,
        }
    return _kops_config


async def post_event(
    *,
    ref: bodies.ObjectReference,
    type: str,
    reason: str,
    message: str = '',
    resource: references.Resource,
    settings: configuration.OperatorSettings,
    logger: typedefs.Logger,
) -> None:
    """
    Issue an event for the object.
    This is where they can also be accumulated, aggregated, grouped,
    and where the rate-limits should be maintained. It can (and should)
    be done by the client library, as it is done in the Go client.
    """

    # Prevent "event explosion", when core v1 events are handled and create other core v1 events.
    # This can happen with `EVERYTHING` without additional filters, or by explicitly serving them.
    conf = _get_config()
    if ref['apiVersion'] == 'v1' and ref['kind'] == 'Event':
        return

    # See #164. For cluster-scoped objects, use the current namespace from the current context.
    # It could be "default", but in some systems, we are limited to one specific namespace only.
    namespace_name: str = ref.get('namespace') or (await api.get_default_namespace()) or 'default'
    namespace = references.NamespaceName(namespace_name)
    full_ref: bodies.ObjectReference = copy.copy(ref)
    full_ref['namespace'] = namespace

    # Prevent a common case of event posting errors but shortening the message.
    if len(message) > conf['MAX_MESSAGE_LENGTH']:
        infix = conf['CUT_MESSAGE_INFIX']
        prefix = message[:conf['MAX_MESSAGE_LENGTH'] // 2 - (len(infix) // 2)]
        suffix = message[-conf['MAX_MESSAGE_LENGTH'] // 2 + (len(infix) - len(infix) // 2):]
        message = f'{prefix}{infix}{suffix}'

    now = datetime.datetime.utcnow()
    body = {
        'metadata': {
            'namespace': namespace,
            'generateName': conf['GENERATE_EVENT_NAME'],
        },

        'action': 'Action?',
        'type': type,
        'reason': reason,
        'message': message,

        'reportingComponent': 'kops',
        'reportingInstance': 'dev',
        'source' : {'component': 'kops'},  # used in the "From" column in `kubectl describe`.

        'involvedObject': full_ref,

        'firstTimestamp': now.isoformat() + 'Z',  # '2019-01-28T18:25:03.000000Z' -- seen in `kubectl describe ...`
        'lastTimestamp': now.isoformat() + 'Z',  # '2019-01-28T18:25:03.000000Z' - seen in `kubectl get events`
        'eventTime': now.isoformat() + 'Z',  # '2019-01-28T18:25:03.000000Z'
    }

    try:
        await api.post(
            url=resource.get_url(namespace=namespace),
            headers={'Content-Type': 'application/json'},
            payload=body,
            logger=logger,
            settings=settings,
        )

    # Events are helpful but auxiliary, they should not fail the handling cycle.
    # Yet we want to notice that something went wrong (in logs).
    except errors.APIError as e:
        logger.warning(f"Failed to post an event. Ignoring and continuing. "
                       f"Code: {e.code}. Message: {e.message}. Details: {e.details}"
                       f"Event: type={type!r}, reason={reason!r}, message={message!r}.")
    except aiohttp.ClientResponseError as e:
        logger.warning(f"Failed to post an event. Ignoring and continuing. "
                       f"Status: {e.status}. Message: {e.message}. "
                       f"Event: type={type!r}, reason={reason!r}, message={message!r}.")
    except aiohttp.ServerDisconnectedError as e:
        logger.warning(f"Failed to post an event. Ignoring and continuing. "
                       f"Message: {e.message}. "
                       f"Event: type={type!r}, reason={reason!r}, message={message!r}.")
    except aiohttp.ClientOSError:
        logger.warning(f"Failed to post an event. Ignoring and continuing. "
                       f"Event: type={type!r}, reason={reason!r}, message={message!r}.")


kopf._cogs.clients.events.post_event = post_event


from __future__ import annotations

from typing import Sequence

from lazyops.libs.dbinit.base import TextClause, text

from lazyops.libs.dbinit.data_structures.grant_to import GrantTo
from lazyops.libs.dbinit.data_structures.privileges import Privilege
from lazyops.libs.dbinit.entities.database import Database
from lazyops.libs.dbinit.entities.database_entity import DatabaseSqlEntity
from lazyops.libs.dbinit.entities.entity import Entity
from lazyops.libs.dbinit.entities.role import Role
from lazyops.libs.dbinit.mixins.grantable import Grantable


class Schema(DatabaseSqlEntity, Grantable):
    """
    Represents a Postgres `Schema <https://www.postgresql.org/docs/15/ddl-schemas.html>`_.
    """

    def __init__(
        self,
        name: str,
        database: Database,
        depends_on: Sequence[Entity] | None = None,
        check_if_exists: bool | None = None,
        owner: Role | None = None,
        grants: Sequence[GrantTo] | None = None,
    ):
        """
        All __init__ params correspond to CREATE SCHEMA arguments and options, see
        `official Postgres documentation <https://www.postgresql.org/docs/current/sql-createschema.html>`_.


        :param name: Unique name of the entity. Must be unique within a database.
        :param database: The :class:`lazyops.libs.dbinit.entities.Database` that this entity belongs to.
        :param depends_on: Any entities that should be created before this one.
        :param check_if_exists: Flag to set existence check behavior. If `True`, will raise an exception during _safe_create if the entity already exists, and will raise an exception during _safe_drop if the entity does not exist.
        :param owner: The :class:`lazyops.libs.dbinit.entitites.Role` who will own this database. Postgres defaults to the user executing the command.
        :param grants: Sequence of :class:`lazyops.libs.dbinit.data_structures.GrantTo` to specify privileges this database has in relation to specified roles.
        """
        self.owner = owner

        Grantable.__init__(self, name=name, grants=grants)
        DatabaseSqlEntity.__init__(
            self, name=name, depends_on=depends_on, database=database, check_if_exists=check_if_exists
        )

    def _create_statements(self) -> Sequence[TextClause]:
        # statement = f"CREATE SCHEMA {self.name}"
        statement = f"CREATE SCHEMA {self.encoded_name}"

        if self.owner:
            statement = f"{statement} AUTHORIZATION {self.owner.encoded_name}"
            # statement = f"{statement} AUTHORIZATION {self.owner.name}"

        return [text(statement)]

    def _exists_statement(self) -> TextClause:
        return text("SELECT EXISTS(SELECT 1 FROM pg_namespace WHERE nspname=:schema)").bindparams(schema=self.name)

    def _drop_statements(self) -> Sequence[TextClause]:
        return [text(f"DROP SCHEMA {self.encoded_name}")]
        # return [text(f"DROP SCHEMA {self.name}")]

    def _grant(self, grantee: Role, privileges: set[Privilege]) -> None:
        self._commit_sql(
            engine=self.database.db_engine(), statements=self._grant_statements(grantee=grantee, privileges=privileges)
        )

    def _grants_exist(self, grantee: Role, privileges: set[Privilege]) -> bool:
        rows = self._fetch_sql(engine=self.database.db_engine(), statement=self._grants_exist_statement())

        try:
            existing_privileges = list(
                filter(None, [self._extract_privileges(acl=r[0], grantee=grantee) for r in rows])
            )[0]
            return privileges.issubset(existing_privileges)
        except IndexError:
            return False

    def _grants_exist_statement(self) -> TextClause:
        """
        The SQL statement that checks to see what grants exist.
        :return: A single :class:`sqlalchemy.TextClause` containing the SQL to check what grants exist on this entity.
        """
        return text("SELECT unnest(nspacl) FROM pg_catalog.pg_namespace WHERE nspname=:schema_name").bindparams(
            schema_name=self.name
        )

    def _revoke(self, grantee: Role, privileges: set[Privilege]) -> None:
        self._commit_sql(
            engine=self.database.db_engine(), statements=self._revoke_statements(grantee=grantee, privileges=privileges)
        )

    @staticmethod
    def _allowed_privileges() -> set[Privilege]:
        return {Privilege.CREATE, Privilege.USAGE}





from lazyops.libs.dbinit.base import create_engine
from lazyops.libs.dbinit.controller import Controller
from lazyops.libs.dbinit.data_structures import GrantOn, GrantTo, Privilege
from lazyops.libs.dbinit.entities import Database, DatabaseContent, Role, Schema


from __future__ import annotations

"""
Base Compression Classes
"""

import abc

from lazyops.utils.logs import logger
# from lazyops.utils.pooler import ThreadPoolV2 as ThreadPooler
from lazyops.utils.pooler import ThreadPooler
from typing import Any, Optional, Union, Dict, TypeVar


class BaseCompression(abc.ABC):
    """
    The Base Compression Class
    """
    name: Optional[str] = None
    compression_level: Optional[int] = None
    encoding: Optional[str] = None

    def __init__(
        self,    
        compression_level: Optional[int] = None,
        encoding: Optional[str] = None,
        raise_errors: bool = False,
        compression_kwargs: Optional[Dict[str, Any]] = None,
        decompression_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs,
    ):
        """
        Initializes the serializer
        """
        if compression_level is not None:
            self.compression_level = compression_level
        if encoding is not None:
            self.encoding = encoding
        self.raise_errors = raise_errors
        self._compression_kwargs = compression_kwargs or {}
        self._decompression_kwargs = decompression_kwargs or {}
        self._kwargs = kwargs
        self.check_deps()
        self.validate_compression_level()

    def validate_compression_level(self):
        """
        Validates the compression level
        """
        pass

    def check_deps(self):
        """
        Checks for dependencies
        """
        pass

    def compress(self, data: Union[str, bytes], level: Optional[int] = None, **kwargs) -> bytes:
        """
        Base Compress
        """
        raise NotImplementedError()
    
    def decompress(self, data: Union[str, bytes], **kwargs) -> Union[str, bytes]:
        """
        Base Decompress
        """
        raise NotImplementedError()
    
    async def acompress(self, data: Union[str, bytes], level: Optional[int] = None, **kwargs) -> bytes:
        """
        Base Compress
        """
        return await ThreadPooler.run_async(self.compress, data, level = level, **kwargs)
    
    async def adecompress(self, data: Union[str, bytes], **kwargs) -> Union[str, bytes]:
        """
        Base Decompress
        """
        return await ThreadPooler.run_async(self.decompress, data, **kwargs)





from lazyops.libs.dbinit.entities.entity import Entity
from lazyops.libs.dbinit.mixins.sql import SQLCreatable


class ClusterEntity(SQLCreatable, Entity):
    """
    Parent class for creatable, cluster-wide entities (like a database or role). Defines some of
    the abstract methods needed since they are consistent across entities.
    """

    def _create(self) -> None:
        self._commit_sql(engine=self.__class__.engine(), statements=self._create_statements())

    def _exists(self) -> bool:
        rows = self._fetch_sql(engine=self.__class__.engine(), statement=self._exists_statement())
        return rows[0][0]  # type: ignore

    def _drop(self) -> None:
        self._commit_sql(engine=self.__class__.engine(), statements=self._drop_statements())



"""
Modified version of 
`googlesearch` by https://github.com/MarioVilas/googlesearch
"""


import abc
import time
import random
import asyncio
import contextlib
from pydantic import BaseModel
from lazyops.imports._aiohttpx import resolve_aiohttpx
from lazyops.imports._bs4 import resolve_bs4
from urllib.parse import quote_plus


resolve_aiohttpx(True)
resolve_bs4(True)

import aiohttpx
from bs4 import BeautifulSoup, Tag
from .utils import filter_result, load_user_agents, load_cookie_jar, get_random_jitter
from typing import List, Optional, Dict, Any, Union, Tuple, Set, Callable, Awaitable, TypeVar, Generator, AsyncGenerator



# URL templates to make Google searches.
url_parameters = (
    'hl', 'q', 'num', 'btnG', 'start', 'tbs', 'safe', 'cr', 'filter'
)

# RT = TypeVar('RT', str, Tuple[str, str])
RT = Union[str, Tuple[str, str], Tuple[str, str, str]]

class SearchParams(BaseModel):
    """
    The search params
    """
    query: str
    tld: Optional[str] = 'com'
    lang: Optional[str] = 'en' 
    tbs: Optional[str] = '0'
    safe: Optional[str] = 'off'
    num: Optional[int] = 10
    start: Optional[int] = 0
    stop: Optional[int] = None
    # pause: Optional[float] = 2.0
    country: Optional[str] = ''
    include_title: Optional[bool] = False
    include_description: Optional[bool] = False
    # extra_params: Optional[Dict[str,str]] = None

    @property
    def url_home(self) -> str:
        """
        Gets the home url
        """
        return f"https://www.google.{self.tld}/"
    
    @property
    def url_search(self) -> str:
        """
        Gets the search url
        """
        return f"https://www.google.{self.tld}/search?lr=lang_{self.lang}&" \
             f"q={self.query}&btnG=Google+Search&tbs={self.tbs}&safe={self.safe}&" \
             f"cr={self.country}&filter=0"
    
    @property
    def url_next_page(self) -> str:
        """
        Gets the next page url
        """
        return f"https://www.google.{self.tld}/search?lr=lang_{self.lang}&" \
                f"q={self.query}&start={self.start}&tbs={self.tbs}&safe={self.safe}&" \
                f"cr={self.country}&filter=0"
    
    @property
    def url_search_num(self) -> str:
        """
        Gets the search url
        """
        return f"https://www.google.{self.tld}/search?lr=lang_{self.lang}&" \
                 f"q={self.query}&num={self.num}&btnG=Google+Search&tbs={self.tbs}&" \
                 f"&safe={self.safe}scr={self.country}&filter=0"
    
    @property
    def url_next_page_num(self) -> str:
        """
        Gets the next page url
        """
        return f"https://www.google.{self.tld}/search?lr=lang_{self.lang}&" \
                    f"q={self.query}&num={self.num}&start={self.start}&tbs={self.tbs}&" \
                    f"safe={self.safe}&cr={self.country}&filter=0"
    
    @property
    def start_url(self) -> str:
        """
        Gets the start url
        """
        if self.start:
            return self.url_next_page if self.num == 10 else self.url_next_page_num
        return self.url_search if self.num == 10 else self.url_search_num
    
    def get_next_url(self) -> str:
        """
        Gets the next url
        """
        self.start += self.num
        return self.url_next_page if self.num == 10 else self.url_next_page_num
    
    def is_valid_link(self, anchor: Tag, hashes: Set) -> Optional[str]:
        """
        Validates the link
        """
        try:
            link = anchor['href']
        except KeyError:
            return None

        # Filter invalid links and links pointing to Google itself.
        link = filter_result(link)
        if not link: return None

        # Discard repeated results.
        h = hash(link)
        if h in hashes: return None
        hashes.add(h)
        return link
    
    def extract_result(self, anchor: Tag, desc: Tag, hashes: Set) -> Optional[RT]:
        """
        Extracts the result
        """
        link = self.is_valid_link(anchor, hashes)
        if not link: return None
        if self.include_title and self.include_description:
            return (link, anchor.text, desc.text)
        if self.include_description:
            return (link, desc.text)
        return (link, anchor.text) if self.include_title else link
    


class GoogleSearchClient(abc.ABC):

    def __init__(
        self,
        proxy: Optional[str] = None,
        user_agents: Optional[List[str]] = None,
        verify_ssl: Optional[bool] = True,
        default_lang: Optional[str] = 'en',
        timeout: Optional[int] = 10,
        max_connections: Optional[int] = 100,
        raise_exceptions: Optional[bool] = True,
        **kwargs,
    ):
        """
        Initializes the client
        """
        self.pre_init(**kwargs)
        self.proxy = proxy
        self.user_agents = user_agents or load_user_agents()
        self.verify_ssl = verify_ssl
        self.lang = default_lang
        self.raise_exceptions = raise_exceptions
        self.cookie_jar = load_cookie_jar()
        self.cookies = aiohttpx.Cookies(self.cookie_jar)
        self.timeout = timeout
        self.max_connections = max_connections
        self.api: aiohttpx.Client = None
        self.init_api(**kwargs)
        self.post_init(**kwargs)

    def pre_init(self, **kwargs):
        """
        Pre init
        """
        pass

    def post_init(self, **kwargs):
        """
        Post init
        """
        pass

    def init_api(self, **kwargs):
        """
        Initializes the api
        """
        if self.api is None:
            self.api = aiohttpx.Client(
                timeout = self.timeout,
                cookies = self.cookies,
                limits = aiohttpx.Limits(max_connections = self.max_connections),
                proxies = {"all://": self.proxy} if self.proxy else None,
                follow_redirects = True,
            )

    
    def _get_page(
        self,
        url: str,
        user_agent: Optional[str] = None,
        **kwargs,
    ) -> bytes:
        """
        Gets the page
        """
        user_agent = user_agent or random.choice(self.user_agents)
        request = self.api.build_request(
            "GET", 
            url, 
        )
        request.headers['User-Agent'] = user_agent
        response = self.api.send(request)
        if self.raise_exceptions: response.raise_for_status()
        self.cookies.extract_cookies(response)
        html = response.read()
        response.close()
        with contextlib.suppress(Exception):
            self.cookies.jar.save()
        return html
    
    def get_page(
        self,
        url: str,
        user_agent: Optional[str] = None,
        **kwargs,
    ) -> bytes:
        """
        Gets the page

        This is a wrapper that allows you to customize the 
        underlying logic of the get_page method
        """
        return self._get_page(url, user_agent, **kwargs)
    
    async def _aget_page(
        self,
        url: str,
        user_agent: Optional[str] = None,
        **kwargs,
    ) -> bytes:
        """
        Gets the page
        """
        user_agent = user_agent or random.choice(self.user_agents)
        request = await self.api.async_build_request(
            "GET", 
            url, 
        )
        request.headers['User-Agent'] = user_agent
        response = await self.api.async_send(request)
        if self.raise_exceptions: response.raise_for_status()
        self.cookies.extract_cookies(response)
        html = response.read()
        await response.aclose()
        with contextlib.suppress(Exception):
            self.cookies.jar.save()
        return html
    
    async def aget_page(
        self,
        url: str,
        user_agent: Optional[str] = None,
        **kwargs,
    ) -> bytes:
        """
        Gets the page

        This is a wrapper that allows you to customize the 
        underlying logic of the get_page method
        """
        return await self._aget_page(url, user_agent, **kwargs)
    

    
    def append_extra_params(
        self,
        url: str,
        extra_params: Dict[str, str],
    ) -> str:
        """
        Appends extra params
        """
        if not extra_params: return url
        for k, v in extra_params.items():
            k = quote_plus(k)
            v = quote_plus(v)
            url = f'{url}&{k}={v}'
        return url
    
    def extract_anchors(
        self,
        html: Union[str, bytes, BeautifulSoup],
    ) -> List[Tag]:
        """
        Extracts the anchors
        """
        soup = BeautifulSoup(html, 'html.parser') if isinstance(html, (bytes, str)) else html
        try:
            return soup.find(id='search').findAll('a')
            # Sometimes (depending on the User-agent) there is
            # no id "search" in html response...
        except AttributeError:
            # Remove links of the top bar.
            gbar = soup.find(id='gbar')
            if gbar: gbar.clear()
            return soup.findAll('a')
        
    def extract_description(
        self,
        html: Union[str, bytes, BeautifulSoup],
    ) -> List[Tag]:
        """
        Extracts the description
        """
        soup = BeautifulSoup(html, 'html.parser') if isinstance(html, (bytes, str)) else html
        try:
            return soup.find(id='search').findAll('span')
        except AttributeError:
            # Remove links of the top bar.
            gbar = soup.find(id='gbar')
            if gbar: gbar.clear()
            return soup.findAll('span')
    
    def validate_extra_params(
        self,
        extra_params: Dict[str, str],
    ):
        """
        Validates the extra params
        """
        for builtin_param in url_parameters:
            if builtin_param in extra_params:
                raise ValueError(
                    'GET parameter "%s" is overlapping with \
                    the built-in GET parameter',
                    builtin_param
                )
    
    def _search(
        self,
        query: str, 
        tld: Optional[str] = 'com', 
        lang: Optional[str] = 'en', 
        tbs: Optional[str] = '0', 
        safe: Optional[str] = 'off', 
        num: Optional[int] = 10, 
        start: Optional[int] = 0,
        stop: Optional[int] = None, 
        pause: Optional[float] = 2.0, 
        country: Optional[str] = '', 
        include_title: Optional[bool] = False,
        include_description: Optional[bool] = False,
        extra_params: Optional[Dict[str,str]] = None,
        user_agent: Optional[str] = None, 
        verify_ssl: Optional[bool] = True,
        **kwargs,
    ) -> Generator[RT, None, None]:  # sourcery skip: low-code-quality
        """
        Searches the query
        """
        hashes = set()

        # Count the number of links yielded.
        count = 0

        # Prepare the search string.
        query = quote_plus(query)

        # If no extra_params is given, create an empty dictionary.
        # We should avoid using an empty dictionary as a default value
        # in a function parameter in Python.
        if not extra_params: extra_params = {}
        if extra_params: self.validate_extra_params(extra_params)

        sp = SearchParams(
            query = query,
            tld = tld,
            lang = lang,
            tbs = tbs,
            safe = safe,
            num = num,
            start = start,
            stop = stop,
            country = country,
            include_title = include_title,
            include_description = include_description,
        )

        # Grab the cookie from the home page.
        self.get_page(sp.url_home, user_agent, **kwargs)

        url = sp.start_url

        # Loop until we reach the maximum result, if any (otherwise, loop forever).
        while not sp.stop or count < sp.stop:

            # Remember last count to detect the end of results.
            last_count = count

            # Append extra GET parameters to the URL.
            # This is done on every iteration because we're
            # rebuilding the entire URL at the end of this loop.
            url = self.append_extra_params(url, extra_params)

            # Sleep between requests.
            # Keeps Google from banning you for making too many requests.
            time.sleep(get_random_jitter(pause))

            # Request the Google Search results page.
            html = self.get_page(url, user_agent)
            soup = BeautifulSoup(html, 'html.parser')

            # Parse the response and get every anchored URL.
            anchors = self.extract_anchors(soup)
            desc_anchors = self.extract_description(soup)

            # Process every anchored URL.
            for (a, desc) in zip(anchors, desc_anchors):
                result = sp.extract_result(a, desc, hashes)
                if not result: continue

                # Yield the result.
                yield result

                # Increase the results counter.
                # If we reached the limit, stop.
                count += 1
                if sp.stop and count >= sp.stop:
                    return

            # End if there are no more results.
            # XXX TODO review this logic, not sure if this is still true!
            if last_count == count:
                break

            # Prepare the URL for the next request.
            url = sp.get_next_url()


    async def _asearch(
        self,
        query: str, 
        tld: Optional[str] = 'com', 
        lang: Optional[str] = 'en', 
        tbs: Optional[str] = '0', 
        safe: Optional[str] = 'off', 
        num: Optional[int] = 10, 
        start: Optional[int] = 0,
        stop: Optional[int] = None, 
        pause: Optional[float] = 2.0, 
        country: Optional[str] = '', 
        include_title: Optional[bool] = False,
        include_description: Optional[bool] = False,
        extra_params: Optional[Dict[str,str]] = None,
        user_agent: Optional[str] = None, 
        verify_ssl: Optional[bool] = True,
        **kwargs,
    ) -> AsyncGenerator[RT, None]:  # sourcery skip: low-code-quality
        """
        Searches the query
        """
        hashes = set()

        # Count the number of links yielded.
        count = 0

        # Prepare the search string.
        query = quote_plus(query)

        # If no extra_params is given, create an empty dictionary.
        # We should avoid using an empty dictionary as a default value
        # in a function parameter in Python.
        if not extra_params: extra_params = {}
        if extra_params: self.validate_extra_params(extra_params)

        sp = SearchParams(
            query = query,
            tld = tld,
            lang = lang,
            tbs = tbs,
            safe = safe,
            num = num,
            start = start,
            stop = stop,
            country = country,
            include_title = include_title,
            include_description = include_description,
        )

        # Grab the cookie from the home page.
        await self.aget_page(sp.url_home, user_agent, **kwargs)

        url = sp.start_url

        # Loop until we reach the maximum result, if any (otherwise, loop forever).
        while not sp.stop or count < sp.stop:

            # Remember last count to detect the end of results.
            last_count = count

            # Append extra GET parameters to the URL.
            # This is done on every iteration because we're
            # rebuilding the entire URL at the end of this loop.
            url = self.append_extra_params(url, extra_params)

            # Sleep between requests.
            # Keeps Google from banning you for making too many requests.
            await asyncio.sleep(get_random_jitter(pause))

            # Request the Google Search results page.
            html = await self.aget_page(url, user_agent, **kwargs)
            soup = BeautifulSoup(html, 'html.parser')

            # Parse the response and get every anchored URL.
            anchors = self.extract_anchors(soup)
            desc_anchors = self.extract_description(soup)
            # Process every anchored URL.
            for (a, desc) in zip(anchors, desc_anchors):
                
                result = sp.extract_result(a, desc, hashes)
                if not result: continue

                # Yield the result.
                yield result

                # Increase the results counter.
                # If we reached the limit, stop.
                count += 1
                if sp.stop and count >= sp.stop: return

            # End if there are no more results.
            # XXX TODO review this logic, not sure if this is still true!
            if last_count == count: break

            # Prepare the URL for the next request.
            url = sp.get_next_url()


    def search(
        self,
        query: str, 
        tld: Optional[str] = 'com', 
        lang: Optional[str] = 'en', 
        tbs: Optional[str] = '0', 
        safe: Optional[str] = 'off', 
        num: Optional[int] = 10, 
        start: Optional[int] = 0,
        stop: Optional[int] = None, 
        pause: Optional[float] = 2.0, 
        country: Optional[str] = '', 
        include_title: Optional[bool] = False,
        include_description: Optional[bool] = False,
        extra_params: Optional[Dict[str,str]] = None,
        user_agent: Optional[str] = None, 
        verify_ssl: Optional[bool] = True,
        **kwargs,
    ) -> Generator[RT, None, None]:  # sourcery skip: low-code-quality
        """
        Searches the query
        """
        yield from self._search(
            query = query, 
            tld = tld, 
            lang = lang, 
            tbs = tbs, 
            safe = safe, 
            num = num, 
            start = start,
            stop = stop, 
            pause = pause, 
            country = country, 
            include_title = include_title,
            include_description = include_description,
            extra_params = extra_params,
            user_agent = user_agent, 
            verify_ssl = verify_ssl,
            **kwargs,
        )

    async def asearch(
        self,
        query: str, 
        tld: Optional[str] = 'com', 
        lang: Optional[str] = 'en', 
        tbs: Optional[str] = '0', 
        safe: Optional[str] = 'off', 
        num: Optional[int] = 10, 
        start: Optional[int] = 0,
        stop: Optional[int] = None, 
        pause: Optional[float] = 2.0, 
        country: Optional[str] = '', 
        include_title: Optional[bool] = False,
        include_description: Optional[bool] = False,
        extra_params: Optional[Dict[str,str]] = None,
        user_agent: Optional[str] = None, 
        verify_ssl: Optional[bool] = True,
        **kwargs,
    ) -> AsyncGenerator[RT, None]:  # sourcery skip: low-code-quality
        """
        Searches the query
        """
        async for item in self._asearch(
            query = query, 
            tld = tld, 
            lang = lang, 
            tbs = tbs, 
            safe = safe, 
            num = num, 
            start = start,
            stop = stop, 
            pause = pause, 
            country = country, 
            include_title = include_title,
            include_description = include_description,
            extra_params = extra_params,
            user_agent = user_agent, 
            verify_ssl = verify_ssl,
            **kwargs,
        ):
            yield item




from __future__ import annotations

"""
FastAPI Utilities
"""

from .configs import (
    WorkerSettings
)

from .openapi import (
    OpenAPIStoplight,
    get_server_domain,
    patch_openapi_schema,
    create_openapi_schema_patch, 
    create_openapi_schema_by_role_function,
    UserRole,
    OpenAPIRoleSpec,
)

from .processes import (
    spawn_new_worker,
    stop_worker,
    run_until_complete,
    arun_until_complete,
    GlobalContext,
    GracefulKiller,
)

from .tasks import (
    register_server_task,
    start_bg_tasks,
    create_server_task,
)

from .utils import (
    create_function_wrapper
)




class PostgresDeclareError(Exception):
    pass


class NoEngineError(PostgresDeclareError):
    pass


class EntityExistsError(PostgresDeclareError):
    pass


class InvalidPrivilegeError(PostgresDeclareError):
    pass



from __future__ import annotations

from abc import ABC, abstractmethod
from collections import defaultdict
from typing import Union, List, Dict, Optional, Tuple, Any
try:
    from sqlalchemy.sql import select
    from sqlalchemy import and_, or_
except ImportError:
    select, and_, or_ = Any, Any, Any

from .utils import convert_date_to_rfc3339


def nested_defaultdict() -> defaultdict:
    """
    Data structure that recursively adds a dictionary as value if a key does not exist. Advantage: In nested dictionary
    structures, we don't need to check if a key already exists (which can become hard to maintain in nested dictionaries
    with many levels) but access the existing value if a key exists and create an empty dictionary if a key does not
    exist.
    """
    return defaultdict(nested_defaultdict)


class LogicalFilterClause(ABC):
    """
    Class that is able to parse a filter and convert it to the format that the underlying databases of our
    DocumentStores require.

    Filters are defined as nested dictionaries. The keys of the dictionaries can be a logical
    operator (`"$and"`, `"$or"`, `"$not"`), a comparison operator (`"$eq"`, `"$in"`, `"$gt"`, `"$gte"`, `"$lt"`,
    `"$lte"`) or a metadata field name.
    Logical operator keys take a dictionary of metadata field names and/or logical operators as
    value. Metadata field names take a dictionary of comparison operators as value. Comparison
    operator keys take a single value or (in case of `"$in"`) a list of values as value.
    If no logical operator is provided, `"$and"` is used as default operation. If no comparison
    operator is provided, `"$eq"` (or `"$in"` if the comparison value is a list) is used as default
    operation.
    Example:
        ```python
        filters = {
            "$and": {
                "type": {"$eq": "article"},
                "date": {"$gte": "2015-01-01", "$lt": "2021-01-01"},
                "rating": {"$gte": 3},
                "$or": {
                    "genre": {"$in": ["economy", "politics"]},
                    "publisher": {"$eq": "nytimes"}
                }
            }
        }
        # or simpler using default operators
        filters = {
            "type": "article",
            "date": {"$gte": "2015-01-01", "$lt": "2021-01-01"},
            "rating": {"$gte": 3},
            "$or": {
                "genre": ["economy", "politics"],
                "publisher": "nytimes"
            }
        }
        ```

    To use the same logical operator multiple times on the same level, logical operators take optionally a list of
    dictionaries as value.

    Example:
        ```python
        filters = {
            "$or": [
                {
                    "$and": {
                        "Type": "News Paper",
                        "Date": {
                            "$lt": "2019-01-01"
                        }
                    }
                },
                {
                    "$and": {
                        "Type": "Blog Post",
                        "Date": {
                            "$gte": "2019-01-01"
                        }
                    }
                }
            ]
        }
        ```

    """

    def __init__(self, conditions: List[Union["LogicalFilterClause", "ComparisonOperation"]]):
        self.conditions = conditions

    @abstractmethod
    def evaluate(self, fields) -> bool:
        pass

    @classmethod
    def parse(cls, filter_term: Union[dict, List[dict]]) -> Union["LogicalFilterClause", "ComparisonOperation"]:
        """
        Parses a filter dictionary/list and returns a LogicalFilterClause instance.

        :param filter_term: Dictionary or list that contains the filter definition.
        """
        conditions: List[Union[LogicalFilterClause, ComparisonOperation]] = []

        if isinstance(filter_term, dict):
            filter_term = [filter_term]
        for item in filter_term:
            for key, value in item.items():
                if key == "$not":
                    conditions.append(NotOperation.parse(value))
                elif key == "$and":
                    conditions.append(AndOperation.parse(value))
                elif key == "$or":
                    conditions.append(OrOperation.parse(value))
                # Key needs to be a metadata field
                else:
                    conditions.extend(ComparisonOperation.parse(key, value))

        if cls == LogicalFilterClause:
            return conditions[0] if len(conditions) == 1 else AndOperation(conditions)
        else:
            return cls(conditions)

    @abstractmethod
    def convert_to_elasticsearch(self):
        """
        Converts the LogicalFilterClause instance to an Elasticsearch filter.
        """
        pass

    @abstractmethod
    def convert_to_sql(self, meta_document_orm):
        """
        Converts the LogicalFilterClause instance to an SQL filter.
        """
        pass

    def convert_to_weaviate(self):
        """
        Converts the LogicalFilterClause instance to a Weaviate filter.
        """
        pass

    def convert_to_pinecone(self):
        """
        Converts the LogicalFilterClause instance to a Pinecone filter.
        """
        pass

    def _merge_es_range_queries(self, conditions: List[Dict]) -> List[Dict[str, Dict]]:
        """
        Merges Elasticsearch range queries that perform on the same metadata field.
        """

        range_conditions = [cond["range"] for cond in filter(lambda condition: "range" in condition, conditions)]
        if range_conditions:
            conditions = [condition for condition in conditions if "range" not in condition]
            range_conditions_dict = nested_defaultdict()
            for condition in range_conditions:
                field_name = list(condition.keys())[0]
                operation = list(condition[field_name].keys())[0]
                comparison_value = condition[field_name][operation]
                range_conditions_dict[field_name][operation] = comparison_value

            conditions.extend(
                {"range": {field_name: comparison_operations}}
                for field_name, comparison_operations in range_conditions_dict.items()
            )
        return conditions

    @abstractmethod
    def invert(self) -> Union["LogicalFilterClause", "ComparisonOperation"]:
        """
        Inverts the LogicalOperation instance.
        Necessary for Weaviate as Weaviate doesn't seem to support the 'Not' operator anymore.
        (https://github.com/semi-technologies/weaviate/issues/1717)
        """
        pass


class ComparisonOperation(ABC):
    def __init__(self, field_name: str, comparison_value: Union[str, int, float, bool, List]):
        self.field_name = field_name
        self.comparison_value = comparison_value

    @abstractmethod
    def evaluate(self, fields) -> bool:
        pass

    @classmethod
    def parse(cls, field_name, comparison_clause: Union[Dict, List, str, float]) -> List["ComparisonOperation"]:
        comparison_operations: List[ComparisonOperation] = []

        if isinstance(comparison_clause, dict):
            for comparison_operation, comparison_value in comparison_clause.items():
                if comparison_operation == "$eq":
                    comparison_operations.append(EqOperation(field_name, comparison_value))
                elif comparison_operation == "$in":
                    comparison_operations.append(InOperation(field_name, comparison_value))
                elif comparison_operation == "$ne":
                    comparison_operations.append(NeOperation(field_name, comparison_value))
                elif comparison_operation == "$nin":
                    comparison_operations.append(NinOperation(field_name, comparison_value))
                elif comparison_operation == "$gt":
                    comparison_operations.append(GtOperation(field_name, comparison_value))
                elif comparison_operation == "$gte":
                    comparison_operations.append(GteOperation(field_name, comparison_value))
                elif comparison_operation == "$lt":
                    comparison_operations.append(LtOperation(field_name, comparison_value))
                elif comparison_operation == "$lte":
                    comparison_operations.append(LteOperation(field_name, comparison_value))

        # No comparison operator is given, so we use the default operators "$in" if the comparison value is a list and
        # "$eq" in every other case
        elif isinstance(comparison_clause, list):
            comparison_operations.append(InOperation(field_name, comparison_clause))
        else:
            comparison_operations.append((EqOperation(field_name, comparison_clause)))

        return comparison_operations

    @abstractmethod
    def convert_to_elasticsearch(self):
        """
        Converts the ComparisonOperation instance to an Elasticsearch query.
        """
        pass

    @abstractmethod
    def convert_to_sql(self, meta_document_orm):
        """
        Converts the ComparisonOperation instance to an SQL filter.
        """
        pass

    @abstractmethod
    def convert_to_weaviate(self):
        """
        Converts the ComparisonOperation instance to a Weaviate comparison operator.
        """
        pass

    def convert_to_pinecone(self):
        """
        Converts the ComparisonOperation instance to a Pinecone comparison operator.
        """
        pass

    @abstractmethod
    def invert(self) -> "ComparisonOperation":
        """
        Inverts the ComparisonOperation.
        Necessary for Weaviate as Weaviate doesn't seem to support the 'Not' operator anymore.
        (https://github.com/semi-technologies/weaviate/issues/1717)
        """
        pass

    def _get_weaviate_datatype(
        self, value: Optional[Union[str, int, float, bool]] = None
    ) -> Tuple[str, Union[str, int, float, bool]]:
        """
        Determines the type of the comparison value and converts it to RFC3339 format if it is as date,
        as Weaviate requires dates to be in RFC3339 format including the time and timezone.

        """
        if value is None:
            assert not isinstance(self.comparison_value, list)  # Necessary for mypy
            value = self.comparison_value

        if isinstance(value, str):
            # Check if comparison value is a date
            try:
                value = convert_date_to_rfc3339(value)
                data_type = "valueDate"
            except ValueError:
                data_type = "valueText" if self.field_name == "content" else "valueString"
        elif isinstance(value, int):
            data_type = "valueInt"
        elif isinstance(value, float):
            data_type = "valueNumber"
        elif isinstance(value, bool):
            data_type = "valueBoolean"
        else:
            raise ValueError(
                f"Unsupported data type of comparison value for {self.__class__.__name__}."
                f"Value needs to be of type str, int, float, or bool."
            )

        return data_type, value


class NotOperation(LogicalFilterClause):
    """
    Handles conversion of logical 'NOT' operations.
    """

    def evaluate(self, fields) -> bool:
        return not any(condition.evaluate(fields) for condition in self.conditions)

    def convert_to_elasticsearch(self) -> Dict[str, Dict]:
        conditions = [condition.convert_to_elasticsearch() for condition in self.conditions]
        conditions = self._merge_es_range_queries(conditions)
        return {"bool": {"must_not": conditions}}

    def convert_to_sql(self, meta_document_orm):
        conditions = [
            meta_document_orm.document_id.in_(condition.convert_to_sql(meta_document_orm))
            for condition in self.conditions
        ]
        return select(meta_document_orm.document_id).filter(~or_(*conditions))

    def convert_to_weaviate(self) -> Dict[str, Union[str, int, float, bool, List[Dict]]]:
        conditions = [condition.invert().convert_to_weaviate() for condition in self.conditions]
        if len(conditions) > 1:
            # Conditions in self.conditions are by default combined with AND which becomes OR according to DeMorgan
            return {"operator": "Or", "operands": conditions}
        else:
            return conditions[0]

    def convert_to_pinecone(self) -> Dict[str, Union[str, int, float, bool, List[Dict]]]:
        conditions = [condition.invert().convert_to_pinecone() for condition in self.conditions]
        return {"$or": conditions} if len(conditions) > 1 else conditions[0]

    def invert(self) -> Union[LogicalFilterClause, ComparisonOperation]:
        # This method is called when a "$not" operation is embedded in another "$not" operation. Therefore, we don't
        # invert the operations here, as two "$not" operation annihilate each other.
        # (If we have more than one condition, we return an AndOperation, the default logical operation for combining
        # multiple conditions.)
        if len(self.conditions) > 1:
            return AndOperation(self.conditions)
        else:
            return self.conditions[0]


class AndOperation(LogicalFilterClause):
    """
    Handles conversion of logical 'AND' operations.
    """

    def evaluate(self, fields) -> bool:
        return all(condition.evaluate(fields) for condition in self.conditions)

    def convert_to_elasticsearch(self) -> Dict[str, Dict]:
        conditions = [condition.convert_to_elasticsearch() for condition in self.conditions]
        conditions = self._merge_es_range_queries(conditions)
        return {"bool": {"must": conditions}}

    def convert_to_sql(self, meta_document_orm):
        conditions = [
            meta_document_orm.document_id.in_(condition.convert_to_sql(meta_document_orm))
            for condition in self.conditions
        ]
        return select(meta_document_orm.document_id).filter(and_(*conditions))

    def convert_to_weaviate(self) -> Dict[str, Union[str, List[Dict]]]:
        conditions = [condition.convert_to_weaviate() for condition in self.conditions]
        return {"operator": "And", "operands": conditions}

    def convert_to_pinecone(self) -> Dict[str, Union[str, List[Dict]]]:
        conditions = [condition.convert_to_pinecone() for condition in self.conditions]
        return {"$and": conditions}

    def invert(self) -> "OrOperation":
        return OrOperation([condition.invert() for condition in self.conditions])


class OrOperation(LogicalFilterClause):
    """
    Handles conversion of logical 'OR' operations.
    """

    def evaluate(self, fields) -> bool:
        return any(condition.evaluate(fields) for condition in self.conditions)

    def convert_to_elasticsearch(self) -> Dict[str, Dict]:
        conditions = [condition.convert_to_elasticsearch() for condition in self.conditions]
        conditions = self._merge_es_range_queries(conditions)
        return {"bool": {"should": conditions}}

    def convert_to_sql(self, meta_document_orm):
        conditions = [
            meta_document_orm.document_id.in_(condition.convert_to_sql(meta_document_orm))
            for condition in self.conditions
        ]
        return select(meta_document_orm.document_id).filter(or_(*conditions))

    def convert_to_weaviate(self) -> Dict[str, Union[str, List[Dict]]]:
        conditions = [condition.convert_to_weaviate() for condition in self.conditions]
        return {"operator": "Or", "operands": conditions}

    def convert_to_pinecone(self) -> Dict[str, Union[str, List[Dict]]]:
        conditions = [condition.convert_to_pinecone() for condition in self.conditions]
        return {"$or": conditions}

    def invert(self) -> AndOperation:
        return AndOperation([condition.invert() for condition in self.conditions])


class EqOperation(ComparisonOperation):
    """
    Handles conversion of the '$eq' comparison operation.
    """

    def evaluate(self, fields) -> bool:
        if self.field_name not in fields:
            return False
        return fields[self.field_name] == self.comparison_value

    def convert_to_elasticsearch(self) -> Dict[str, Dict[str, Union[str, int, float, bool]]]:
        assert not isinstance(self.comparison_value, list), "Use '$in' operation for lists as comparison values."
        return {"term": {self.field_name: self.comparison_value}}

    def convert_to_sql(self, meta_document_orm):
        return select([meta_document_orm.document_id]).where(
            meta_document_orm.name == self.field_name, meta_document_orm.value == self.comparison_value
        )

    def convert_to_weaviate(self) -> Dict[str, Union[List[str], str, int, float, bool]]:
        comp_value_type, comp_value = self._get_weaviate_datatype()
        return {"path": [self.field_name], "operator": "Equal", comp_value_type: comp_value}

    def convert_to_pinecone(self) -> Dict[str, Dict[str, Union[List[str], str, int, float, bool]]]:
        return {self.field_name: {"$eq": self.comparison_value}}

    def invert(self) -> "NeOperation":
        return NeOperation(self.field_name, self.comparison_value)


class InOperation(ComparisonOperation):
    """
    Handles conversion of the '$in' comparison operation.
    """

    def evaluate(self, fields) -> bool:
        if self.field_name not in fields:
            return False
        return fields[self.field_name] in self.comparison_value  # type: ignore
        # is only initialized with lists, but changing the type annotation would mean duplicating __init__

    def convert_to_elasticsearch(self) -> Dict[str, Dict[str, List]]:
        assert isinstance(self.comparison_value, list), "'$in' operation requires comparison value to be a list."
        return {"terms": {self.field_name: self.comparison_value}}

    def convert_to_sql(self, meta_document_orm):
        return select([meta_document_orm.document_id]).where(
            meta_document_orm.name == self.field_name, meta_document_orm.value.in_(self.comparison_value)
        )

    def convert_to_weaviate(self) -> Dict[str, Union[str, List[Dict]]]:
        filter_dict: Dict[str, Union[str, List[Dict]]] = {"operator": "Or", "operands": []}
        assert isinstance(self.comparison_value, list), "'$in' operation requires comparison value to be a list."
        for value in self.comparison_value:
            comp_value_type, comp_value = self._get_weaviate_datatype(value)
            assert isinstance(filter_dict["operands"], list)  # Necessary for mypy
            filter_dict["operands"].append(
                {"path": [self.field_name], "operator": "Equal", comp_value_type: comp_value}
            )

        return filter_dict

    def convert_to_pinecone(self) -> Dict[str, Dict[str, List]]:
        assert isinstance(self.comparison_value, list), "'$in' operation requires comparison value to be a list."
        return {self.field_name: {"$in": self.comparison_value}}

    def invert(self) -> "NinOperation":
        return NinOperation(self.field_name, self.comparison_value)


class NeOperation(ComparisonOperation):
    """
    Handles conversion of the '$ne' comparison operation.
    """

    def evaluate(self, fields) -> bool:
        if self.field_name not in fields:
            return False
        return fields[self.field_name] != self.comparison_value

    def convert_to_elasticsearch(self) -> Dict[str, Dict[str, Dict[str, Dict[str, Union[str, int, float, bool]]]]]:
        assert not isinstance(self.comparison_value, list), "Use '$nin' operation for lists as comparison values."
        return {"bool": {"must_not": {"term": {self.field_name: self.comparison_value}}}}

    def convert_to_sql(self, meta_document_orm):
        return select([meta_document_orm.document_id]).where(
            meta_document_orm.name == self.field_name, meta_document_orm.value != self.comparison_value
        )

    def convert_to_weaviate(self) -> Dict[str, Union[List[str], str, int, float, bool]]:
        comp_value_type, comp_value = self._get_weaviate_datatype()
        return {"path": [self.field_name], "operator": "NotEqual", comp_value_type: comp_value}

    def convert_to_pinecone(self) -> Dict[str, Dict[str, Union[List[str], str, int, float, bool]]]:
        return {self.field_name: {"$ne": self.comparison_value}}

    def invert(self) -> "EqOperation":
        return EqOperation(self.field_name, self.comparison_value)


class NinOperation(ComparisonOperation):
    """
    Handles conversion of the '$nin' comparison operation.
    """

    def evaluate(self, fields) -> bool:
        if self.field_name not in fields:
            return False
        return fields[self.field_name] not in self.comparison_value  # type: ignore
        # is only initialized with lists, but changing the type annotation would mean duplicating __init__

    def convert_to_elasticsearch(self) -> Dict[str, Dict[str, Dict[str, Dict[str, List]]]]:
        assert isinstance(self.comparison_value, list), "'$nin' operation requires comparison value to be a list."
        return {"bool": {"must_not": {"terms": {self.field_name: self.comparison_value}}}}

    def convert_to_sql(self, meta_document_orm):
        return select([meta_document_orm.document_id]).where(
            meta_document_orm.name == self.field_name, meta_document_orm.value.notin_(self.comparison_value)
        )

    def convert_to_weaviate(self) -> Dict[str, Union[str, List[Dict]]]:
        filter_dict: Dict[str, Union[str, List[Dict]]] = {"operator": "And", "operands": []}
        assert isinstance(self.comparison_value, list), "'$nin' operation requires comparison value to be a list."
        for value in self.comparison_value:
            comp_value_type, comp_value = self._get_weaviate_datatype(value)
            assert isinstance(filter_dict["operands"], list)  # Necessary for mypy
            filter_dict["operands"].append(
                {"path": [self.field_name], "operator": "NotEqual", comp_value_type: comp_value}
            )

        return filter_dict

    def convert_to_pinecone(self) -> Dict[str, Dict[str, List]]:
        assert isinstance(self.comparison_value, list), "'$in' operation requires comparison value to be a list."
        return {self.field_name: {"$nin": self.comparison_value}}

    def invert(self) -> "InOperation":
        return InOperation(self.field_name, self.comparison_value)


class GtOperation(ComparisonOperation):
    """
    Handles conversion of the '$gt' comparison operation.
    """

    def evaluate(self, fields) -> bool:
        if self.field_name not in fields:
            return False
        return fields[self.field_name] > self.comparison_value

    def convert_to_elasticsearch(self) -> Dict[str, Dict[str, Dict[str, Union[str, float, int]]]]:
        assert not isinstance(self.comparison_value, list), "Comparison value for '$gt' operation must not be a list."
        return {"range": {self.field_name: {"gt": self.comparison_value}}}

    def convert_to_sql(self, meta_document_orm):
        return select([meta_document_orm.document_id]).where(
            meta_document_orm.name == self.field_name, meta_document_orm.value > self.comparison_value
        )

    def convert_to_weaviate(self) -> Dict[str, Union[List[str], str, float, int]]:
        comp_value_type, comp_value = self._get_weaviate_datatype()
        assert not isinstance(comp_value, list), "Comparison value for '$gt' operation must not be a list."
        return {"path": [self.field_name], "operator": "GreaterThan", comp_value_type: comp_value}

    def convert_to_pinecone(self) -> Dict[str, Dict[str, Union[float, int]]]:
        assert not isinstance(
            self.comparison_value, (list, str)
        ), "Comparison value for '$gt' operation must be a float or int."
        return {self.field_name: {"$gt": self.comparison_value}}

    def invert(self) -> "LteOperation":
        return LteOperation(self.field_name, self.comparison_value)


class GteOperation(ComparisonOperation):
    """
    Handles conversion of the '$gte' comparison operation.
    """

    def evaluate(self, fields) -> bool:
        if self.field_name not in fields:
            return False
        return fields[self.field_name] >= self.comparison_value

    def convert_to_elasticsearch(self) -> Dict[str, Dict[str, Dict[str, Union[str, float, int]]]]:
        assert not isinstance(self.comparison_value, list), "Comparison value for '$gte' operation must not be a list."
        return {"range": {self.field_name: {"gte": self.comparison_value}}}

    def convert_to_sql(self, meta_document_orm):
        return select([meta_document_orm.document_id]).where(
            meta_document_orm.name == self.field_name, meta_document_orm.value >= self.comparison_value
        )

    def convert_to_weaviate(self) -> Dict[str, Union[List[str], str, float, int]]:
        comp_value_type, comp_value = self._get_weaviate_datatype()
        assert not isinstance(comp_value, list), "Comparison value for '$gte' operation must not be a list."
        return {"path": [self.field_name], "operator": "GreaterThanEqual", comp_value_type: comp_value}

    def convert_to_pinecone(self) -> Dict[str, Dict[str, Union[float, int]]]:
        assert not isinstance(
            self.comparison_value, (list, str)
        ), "Comparison value for '$gte' operation must be a float or int."
        return {self.field_name: {"$gte": self.comparison_value}}

    def invert(self) -> "LtOperation":
        return LtOperation(self.field_name, self.comparison_value)


class LtOperation(ComparisonOperation):
    """
    Handles conversion of the '$lt' comparison operation.
    """

    def evaluate(self, fields) -> bool:
        if self.field_name not in fields:
            return False
        return fields[self.field_name] < self.comparison_value

    def convert_to_elasticsearch(self) -> Dict[str, Dict[str, Dict[str, Union[str, float, int]]]]:
        assert not isinstance(self.comparison_value, list), "Comparison value for '$lt' operation must not be a list."
        return {"range": {self.field_name: {"lt": self.comparison_value}}}

    def convert_to_sql(self, meta_document_orm):
        return select([meta_document_orm.document_id]).where(
            meta_document_orm.name == self.field_name, meta_document_orm.value < self.comparison_value
        )

    def convert_to_weaviate(self) -> Dict[str, Union[List[str], str, float, int]]:
        comp_value_type, comp_value = self._get_weaviate_datatype()
        assert not isinstance(comp_value, list), "Comparison value for '$lt' operation must not be a list."
        return {"path": [self.field_name], "operator": "LessThan", comp_value_type: comp_value}

    def convert_to_pinecone(self) -> Dict[str, Dict[str, Union[float, int]]]:
        assert not isinstance(
            self.comparison_value, (list, str)
        ), "Comparison value for '$lt' operation must be a float or int."
        return {self.field_name: {"$lt": self.comparison_value}}

    def invert(self) -> "GteOperation":
        return GteOperation(self.field_name, self.comparison_value)


class LteOperation(ComparisonOperation):
    """
    Handles conversion of the '$lte' comparison operation.
    """

    def evaluate(self, fields) -> bool:
        if self.field_name not in fields:
            return False
        return fields[self.field_name] <= self.comparison_value

    def convert_to_elasticsearch(self) -> Dict[str, Dict[str, Dict[str, Union[str, float, int]]]]:
        assert not isinstance(self.comparison_value, list), "Comparison value for '$lte' operation must not be a list."
        return {"range": {self.field_name: {"lte": self.comparison_value}}}

    def convert_to_sql(self, meta_document_orm):
        return select([meta_document_orm.document_id]).where(
            meta_document_orm.name == self.field_name, meta_document_orm.value <= self.comparison_value
        )

    def convert_to_weaviate(self) -> Dict[str, Union[List[str], str, float, int]]:
        comp_value_type, comp_value = self._get_weaviate_datatype()
        assert not isinstance(comp_value, list), "Comparison value for '$lte' operation must not be a list."
        return {"path": [self.field_name], "operator": "LessThanEqual", comp_value_type: comp_value}

    def convert_to_pinecone(self) -> Dict[str, Dict[str, Union[float, int]]]:
        assert not isinstance(
            self.comparison_value, (list, str)
        ), "Comparison value for '$lte' operation must be a float or int."
        return {self.field_name: {"$lte": self.comparison_value}}

    def invert(self) -> "GtOperation":
        return GtOperation(self.field_name, self.comparison_value)



import time
from lazyops.utils.helpers import fail_after

def blocking_func():
    while True:
        print("blocking")
        time.sleep(1)

def test_func():

    try:
        with fail_after(5):
            blocking_func()
    
    except Exception as e:
        print(e)

    print("done")


if __name__ == "__main__":
    test_func()





import os
import gzip
import random
import pathlib
import datetime
import contextlib
from http.cookiejar import LWPCookieJar
from urllib.parse import urlparse, parse_qs
from typing import List, Optional, TYPE_CHECKING
if TYPE_CHECKING:
    from http.cookiejar import CookieJar



lib_path = pathlib.Path(__file__).parent.absolute()

home_folder = os.getenv('HOME', os.getenv('USERHOME'))

# Default user agent, unless instructed by the user to change it.
USER_AGENT = 'Mozilla/4.0 (compatible; MSIE 8.0; Windows NT 6.0)'


# Load the list of valid user agents from the install folder.
# The search order is:
#   * user_agents.txt.gz
#   * user_agents.txt
#   * default user agent

_user_agents: List[str] = None
_cookie_jar: LWPCookieJar = None

def load_cookie_jar() -> LWPCookieJar:
    """
    Load the cookie jar
    """
    global _cookie_jar
    if _cookie_jar is None:
        _cookie_jar = LWPCookieJar(
            os.path.join(
                home_folder or lib_path.as_posix(), 
                '.google-cookie'
            )
        )

        with contextlib.suppress(Exception):
            _cookie_jar.load()
    
    return _cookie_jar

def load_user_agents():
    """
    Loads the user agents from the user agents file.
    """
    global _user_agents
    if _user_agents is None:
        user_agents_file = lib_path.joinpath('user_agents.txt.gz')
        fp = gzip.open(user_agents_file.as_posix(), 'rb')
        try:
            _user_agents = [_.decode().strip() for _ in fp.readlines()]
        except Exception as e:
            _user_agents = [USER_AGENT]
        finally:
            fp.close()
    
    return _user_agents

def get_random_jitter(
    min_seconds: int = 2, 
    max_seconds: int = 5
) -> int:
    """
    Gets a random jitter
    """
    return random.randint(min_seconds, max_seconds)


# Get a random user agent.
def get_random_user_agent():
    """
    Get a random user agent string.

    :rtype: str
    :return: Random user agent string.
    """
    return random.choice(load_user_agents())



# Helper function to format the tbs parameter.
def get_tbs(from_date: datetime.datetime, to_date: datetime.datetime) -> str:
    """
    Helper function to format the tbs parameter.

    :param datetime.date from_date: Python date object.
    :param datetime.date to_date: Python date object.

    :rtype: str
    :return: Dates encoded in tbs format.
    """
    from_date = from_date.strftime('%m/%d/%Y')
    to_date = to_date.strftime('%m/%d/%Y')
    return f'cdr:1,cd_min:{from_date},cd_max:{to_date}'


# Filter links found in the Google result pages HTML code.
# Returns None if the link doesn't yield a valid result.
def filter_result(link: str) -> Optional[str]:
    """
    Filter links found in the Google result pages HTML code.
    """
    with contextlib.suppress(Exception):
        # Decode hidden URLs.
        if link.startswith('/url?'):
            o = urlparse(link, 'http')
            link = parse_qs(o.query)['q'][0]

        # Valid results are absolute URLs not pointing to a Google domain,
        # like images.google.com or googleusercontent.com for example.
        # TODO this could be improved!
        o = urlparse(link, 'http')
        if o.netloc and 'google' not in o.netloc:
            return link
    
    return None



"""
Persistence Types

- Handles persistence of data with Redis or JSON+Pickle
"""
import atexit
import asyncio
import contextlib
import collections.abc
from lazyops.utils.lazy import lazy_import, get_keydb_enabled
from lazyops.utils.logs import logger, null_logger
# from lazyops.utils.pooler import ThreadPoolV2 as ThreadPooler
from lazyops.utils.pooler import ThreadPooler

from typing import Any, Dict, Optional, Union, Iterable, List, Type, TYPE_CHECKING
from .backends import LocalStatefulBackend, RedisStatefulBackend, StatefulBackendT
from .serializers import ObjectValue

if TYPE_CHECKING:
    from lazyops.types.models import BaseSettings

_DEBUG_ENABLED = False
autologger = logger if _DEBUG_ENABLED else null_logger


RegisteredBackends = {
    'local': LocalStatefulBackend,
    'redis': RedisStatefulBackend,
}

def is_in_async_loop() -> bool:
    """
    Returns True if the function is called in an async loop
    """
    with contextlib.suppress(Exception):
        return asyncio.get_event_loop().is_running()
    return False

class PersistentDict(collections.abc.MutableMapping):
    """
    Persistent Dictionary Interface
    """

    backend_type: Optional[str] = 'auto'
    base_class: Optional[Type[StatefulBackendT]] = None

    def __init__(
        self,
        name: Optional[str] = None,
        serializer: Optional[str] = None,
        serializer_kwargs: Optional[Dict[str, Any]] = None,
        base_key: Optional[str] = None,
        async_enabled: Optional[bool] = False,
        settings: Optional['BaseSettings'] = None,
        backend: Optional[StatefulBackendT] = None,
        backend_type: Optional[str] = None,
        **kwargs,
    ):
        """
        Initializes the Persistent Dictionary
        """
        if backend_type is not None: self.backend_type = backend_type
        if backend is not None:
            self.base_class = lazy_import(backend) if isinstance(backend, str) else backend
        if self.base_class is None:
            self.base_class = self.get_backend_class()
        self.name = name
        self.base_key = base_key
        self.settings = settings
        self._kwargs = kwargs
        self._kwargs['serializer'] = serializer
        self._kwargs['serializer_kwargs'] = serializer_kwargs
        self.base = self.base_class(
            name = self.name,
            base_key = self.base_key,
            async_enabled = async_enabled,
            settings = self.settings,
            **self._kwargs,
        )
        self._mutation_tracker: Dict[str, ObjectValue] = {}
        self._mutation_hashes: Dict[str, str] = {}
        atexit.register(self.flush)
    
    @classmethod
    def register_backend(cls, name: str, backend: Union[str, Type[StatefulBackendT]]):
        """
        Registers a Backend
        """
        global RegisteredBackends
        RegisteredBackends[name] = backend

    def get_backend_class(self, **kwargs) -> Type[StatefulBackendT]:
        """
        Returns the Backend Class
        """
        if self.backend_type in RegisteredBackends:
            bt = RegisteredBackends[self.backend_type]
            return lazy_import(bt) if isinstance(bt, str) else bt
        # if self.backend_type == 'local':
        #     return LocalStatefulBackend
        # elif self.backend_type == 'redis':
        #     return RedisStatefulBackend
        if self.backend_type == 'auto':
            if get_keydb_enabled():
                return RedisStatefulBackend
            return LocalStatefulBackend
        raise NotImplementedError(f'Backend Type {self.backend_type} is not implemented')
    
    @property
    def compression_level(self) -> Optional[int]:
        """
        Returns the Compression Level
        """
        return self.base.serializer.compression_level

    def get_child(self, key: str, **kwargs) -> 'PersistentDict':
        """
        Gets a Child Persistent Dictionary
        """
        base_key = f'{self.base_key}.{key}' if self.base_key else key
        if 'async_enabled' not in kwargs:
            kwargs['async_enabled'] = self.base.async_enabled
        base_kwargs = self._kwargs.copy()
        base_kwargs.update(kwargs)
        return self.__class__(base_key = base_key, **base_kwargs)


    def get(self, key: str, default: Optional[Any] = None, **kwargs) -> Optional[Any]:
        """
        Gets a Value from the DB
        """
        self._save_mutation_objects(key)
        return self.base.get(key, default = default, **kwargs)
    
    def get_values(self, keys: Iterable[str]) -> List[Any]:
        """
        Gets a Value from the DB
        """
        self._save_mutation_objects(*keys)
        return self.base.get_values(keys)
    
    def fetch(self, key: str) -> Optional[Any]:
        """
        Gets a Value from the DB
        """
        return self.get(key) if self.contains(key) else None
    
    def set(self, key: str, value: Any, **kwargs) -> None:
        """
        Saves a Value to the DB
        """
        if self.base.async_enabled and is_in_async_loop():
            ThreadPooler.create_background_task(self.base.aset(key, value, **kwargs))
        else:
            self.base.set(key, value, **kwargs)
    
    def set_batch(self, data: Dict[str, Any], **kwargs) -> None:
        """
        Saves a Value to the DB
        """
        if self.base.async_enabled:
            ThreadPooler.create_background_task(self.base.aset_batch(data, **kwargs))
        else:
            self.base.set_batch(data, **kwargs)

    def delete(self, key: str, **kwargs) -> None:
        """
        Deletes a Value from the DB
        """
        if self.base.async_enabled and is_in_async_loop():
            ThreadPooler.create_background_task(self.base.adelete(key, **kwargs))
        else:
            self.base.delete(key, **kwargs)

    def contains(self, key: str, **kwargs) -> bool:
        """
        Returns True if the Cache contains the Key
        """
        return self.base.contains(key, **kwargs)
    
    def clear(self, *keys, **kwargs) -> None:
        """
        Clears the Cache
        """
        if self.base.async_enabled and is_in_async_loop():
            ThreadPooler.create_background_task(self.base.clear(*keys, **kwargs))
        else:
            self.base.clear(*keys, **kwargs)
    
    async def aget(self, key: str, default: Optional[Any] = None, **kwargs) -> Optional[Any]:
        """
        Gets a Value from the DB
        """
        await self._asave_mutation_objects(key)
        return await self.base.aget(key, default = default, **kwargs)
    
    async def aget_values(self, keys: Iterable[str]) -> List[Any]:
        """
        Gets a Value from the DB
        """
        await self._asave_mutation_objects(*keys)
        return await self.base.aget_values(keys)
    
    async def afetch(self, key: str) -> Optional[Any]:
        """
        Gets a Value from the DB
        """
        return await self.aget(key) if await self.acontains(key) else None

    async def aset(self, key: str, value: Any, **kwargs) -> None:
        """
        Saves a Value to the DB
        """
        await self.base.aset(key, value, **kwargs)

    async def aset_batch(self, data: Dict[str, Any], **kwargs) -> None:
        """
        Saves a Value to the DB
        """
        await self.base.aset_batch(data, **kwargs)

    async def adelete(self, key: str, **kwargs) -> None:
        """
        Deletes a Value from the DB
        """
        await self.base.adelete(key, **kwargs)
    
    async def acontains(self, key: str, **kwargs) -> bool:
        """
        Returns True if the Cache contains the Key
        """
        return await self.base.acontains(key, **kwargs)
    
    async def aclear(self, *keys, **kwargs) -> None:
        """
        Clears the Cache
        """
        await self.base.aclear(*keys, **kwargs)
    
    def get_all_data(self) -> Dict[str, Any]:
        """
        Loads all the Data
        """
        self._save_mutation_objects()
        return self.base.get_all_data()
    
    def get_all_keys(self) -> Iterable[str]:
        """
        Returns all the Keys
        """
        return self.base.get_all_keys()
    
    def get_keys(self, pattern: str, exclude_base_key: Optional[bool] = None) -> List[str]:
        """
        Returns all the Keys
        """
        return self.base.get_keys(pattern, exclude_base_key = exclude_base_key)
    
    def get_all_values(self) -> Iterable[Any]:
        """
        Returns all the Values
        """
        self._save_mutation_objects()
        return self.base.get_all_values()
    
    async def aget_all_data(self) -> Dict[str, Any]:
        """
        Loads all the Data
        """
        await self._asave_mutation_objects()
        return await self.base.aget_all_data()
    
    async def aget_all_keys(self) -> Iterable[str]:
        """
        Returns all the Keys
        """
        return await self.base.aget_all_keys()

    async def aget_keys(self, pattern: str, exclude_base_key: Optional[bool] = None) -> List[str]:
        """
        Returns all the Keys
        """
        return await self.base.aget_keys(pattern, exclude_base_key = exclude_base_key)
    
    async def aget_all_values(self) -> Iterable[Any]:
        """
        Returns all the Values
        """
        await self._asave_mutation_objects()
        return await self.base.aget_all_values()
    
    def keys(self) -> Iterable[Any]:
        """
        Returns the Keys
        """
        return self.base.keys()
    
    def values(self) -> Iterable[Any]:
        """
        Returns the Values
        """
        self._save_mutation_objects()
        return self.base.values()
    
    def items(self, iterable: Optional[bool] = True) -> Dict[Any, Any]:
        """
        Returns the Items
        """
        self._save_mutation_objects()
        return self.base.items(iterable = iterable)
    
    async def akeys(self) -> Iterable[Any]:
        """
        Returns the Keys
        """
        return await self.base.akeys()
    
    async def avalues(self) -> Iterable[Any]:
        """
        Returns the Values
        """
        await self._asave_mutation_objects()
        return await self.base.avalues()
    
    async def aitems(self, iterable: Optional[bool] = True) -> Dict[Any, Any]:
        """
        Returns the Items
        """
        await self._asave_mutation_objects()
        return await self.base.aitems(iterable = iterable)
    
    def expire(self, key: str, timeout: int) -> None:
        """
        Expires a Key
        """
        self.base.expire(key, timeout)

    async def aexpire(self, key: str, timeout: int) -> None:
        """
        Expires a Key
        """
        await self.base.aexpire(key, timeout)

    @contextlib.contextmanager
    def track_changes(self, key: str, func: str, *args, **kwargs):
        """
        Tracks Changes
        """
        try:
            value = None
            if key in self._mutation_tracker:
                autologger.info(f'tracked {func} {key} (cached) {self._mutation_tracker[key]}')
                if self.base.create_hash(self._mutation_tracker[key]) == self._mutation_hashes[key]:
                    value = self._mutation_tracker[key]
                else:
                    autologger.info(f'tracked {func} {key} (changed). Saving')
                    self.base.set(key, self._mutation_tracker[key])
                    value = self._mutation_tracker.pop(key)
                    self._mutation_hashes.pop(key)
            
            if value is None:
                autologger.info(f'tracked {func} {key}')
                value = getattr(self.base, func)(key, *args, **kwargs)
            yield value
        finally:
            if key not in self._mutation_hashes:
                self._mutation_hashes[key] = self.base.create_hash(value)
                self._mutation_tracker[key] = value
            if self.base.create_hash(value) != self._mutation_hashes[key]:
                autologger.info(f'tracked {func} {key} (post-changed). Saving')
                self._save_mutation_objects(key)

    
    @contextlib.asynccontextmanager
    async def atrack_changes(self, key: str, func: str, *args, **kwargs):
        """
        Tracks Changes
        """
        try:
            value = None
            if key in self._mutation_tracker:
                autologger.info(f'tracked {func} {key} (cached): {self._mutation_tracker[key]}')
                if self.base.create_hash(self._mutation_tracker[key]) == self._mutation_hashes[key]:
                    value = self._mutation_tracker[key]
                else:
                    autologger.info(f'tracked {func} {key} (changed). Saving')
                    await self.base.aset(key, self._mutation_tracker[key])
                    value = self._mutation_tracker.pop(key)
                    self._mutation_hashes.pop(key)
            if value is None:
                autologger.info(f'tracked {func} {key}')
                value = await getattr(self.base, func)(key, *args, **kwargs)
            yield value
        finally:
            if key not in self._mutation_hashes:
                self._mutation_hashes[key] = self.base.create_hash(value)
                self._mutation_tracker[key] = value
            if self.base.create_hash(value) != self._mutation_hashes[key]:
                autologger.info(f'tracked {func} {key} (post-changed). Saving')
                await self._asave_mutation_objects(key)

    def setdefault(self, key: str, default: Any = None, update_values: Optional[bool] = False, enforce_type: Optional[bool] = False) -> Any:
        """
        Sets a Default Value
        """
        with self.track_changes(key, 'setdefault', default, update_values = update_values, enforce_type = enforce_type) as result:
            return result
    
    async def asetdefault(self, key: str, default: Any = None, update_values: Optional[bool] = False, enforce_type: Optional[bool] = False) -> Any:
        """
        Sets a Default Value
        """
        async with self.atrack_changes(key, 'asetdefault', default, update_values = update_values, enforce_type = enforce_type) as result:
            return result
        
    
    def update(self, data: Dict[str, Any]) -> None:
        """
        Updates the Cache
        """
        self._save_mutation_objects()
        self.base.update(data)

    async def aupdate(self, data: Dict[str, Any]) -> None:
        """
        Updates the Cache
        """
        await self._asave_mutation_objects()
        await self.base.aupdate(data)

    def popitem(self) -> Any:
        """
        Pops an Item from the Cache
        """
        return self.base.popitem()
    
    async def apopitem(self) -> Any:
        """
        Pops an Item from the Cache
        """
        return await self.base.apopitem()
    
    def pop(self, key: str, default: Optional[Any] = None) -> Any:
        """
        Pops an Item from the Cache
        """
        return self.base.pop(key, default)
    
    async def apop(self, key: str, default: Optional[Any] = None) -> Any:
        """
        Pops an Item from the Cache
        """
        return await self.base.apop(key, default)
    
    def __repr__(self):
        """
        Returns the Representation of the Cache
        """
        return repr(self.base)

    def _clear_from_mutation_tracker(self, key: str):
        """
        Clears the Mutation Tracker
        """
        _ = self._mutation_tracker.pop(key, None)
        _ = self._mutation_hashes.pop(key, None)

    def _save_mutation_objects(self, *keys: str):
        """
        Saves the Mutation Objects
        """
        if not self._mutation_tracker: return
        if keys:
            for key in keys:
                if key in self._mutation_tracker:
                    self.base.set(key, self._mutation_tracker[key])
                    self._clear_from_mutation_tracker(key)
        else:
            autologger.info(f'_save_mutation_objects: {list(self._mutation_tracker.keys())}')
            self.base.set_batch(self._mutation_tracker)
            self._mutation_tracker = {}
            self._mutation_hashes = {}

    async def _asave_mutation_objects(self, *keys: str):
        """
        Saves the Mutation Objects
        """
        if not self._mutation_tracker: return
        if keys:
            for key in keys:
                if key in self._mutation_tracker:
                    await self.base.aset(key, self._mutation_tracker[key])
                    self._clear_from_mutation_tracker(key)
        else:
            autologger.info(f'_save_mutation_objects: {list(self._mutation_tracker.keys())}')
            await self.base.aset_batch(self._mutation_tracker)
            self._mutation_tracker = {}
            self._mutation_hashes = {}

    def __getitem__(self, key: str) -> ObjectValue:
        """
        Gets an Item from the DB
        """
        with self.track_changes(key, '__getitem__') as result:
            return result

    
    def __setitem__(self, key: str, value: ObjectValue):
        """
        Sets an Item in the Cache
        """
        autologger.info(f'__setitem__ {key} {value}')
        if key in self._mutation_tracker:
            self._clear_from_mutation_tracker(key)
        return self.base.__setitem__(key, value)
        
    def __delitem__(self, key):
        """
        Deletes an Item from the Cache
        """
        autologger.info(f'__delitem__ {key}')
        if key in self._mutation_tracker:
            self._clear_from_mutation_tracker(key)
        return self.base.__delitem__(key)
        
    def __iter__(self):
        """
        Iterates over the Cache
        """
        return iter(self.base.keys())
    
    def __len__(self):
        """
        Returns the Length of the Cache
        """
        return len(self.base)
    
    def __contains__(self, key: str):
        """
        Returns True if the Cache contains the Key
        """
        return self.base.contains(key)
    
    def __bool__(self):
        """
        Returns True if the Cache is not Empty
        """
        return bool(self.base.keys())
    

    def migrate_compression(self, **kwargs):
        """
        Migrates the compression
        """
        return self.base.migrate_compression(**kwargs)

    async def amigrate_compression(self, **kwargs):
        """
        Migrates the compression
        """
        return await self.base.amigrate_compression(**kwargs)
    
    def flush(self, *keys: str):
        """
        Finalize any in-memory objects
        """
        if self.base.async_enabled and is_in_async_loop():
            ThreadPooler.create_background_task(self._asave_mutation_objects(*keys))
        else:
            self._save_mutation_objects(*keys)

    async def aflush(self, *keys: str):
        """
        Finalize any in-memory objects
        """
        await self._asave_mutation_objects(*keys)
        